"""
This file contains global options for rioxarray

Credits:

This file was adopted from: https://github.com/pydata/xarray # noqa
Source file: https://github.com/pydata/xarray/blob/2ab0666c1fcc493b1e0ebc7db14500c427f8804e/xarray/core/options.py  # noqa
"""
from typing import Any, Optional

from rioxarray.enum import Convention

EXPORT_GRID_MAPPING = "export_grid_mapping"
SKIP_MISSING_SPATIAL_DIMS = "skip_missing_spatial_dims"
CONVENTION = "convention"

OPTIONS: dict[str, Any] = {
    EXPORT_GRID_MAPPING: True,
    SKIP_MISSING_SPATIAL_DIMS: False,
    CONVENTION: None,
}
OPTION_NAMES = set(OPTIONS)


def _validate_convention(value: Optional[Convention]) -> bool:
    """Validate the convention option."""
    return value is None or isinstance(value, Convention)


VALIDATORS = {
    EXPORT_GRID_MAPPING: lambda choice: isinstance(choice, bool),
    CONVENTION: _validate_convention,
}


def get_option(key: str) -> Any:
    """
    Get the global rioxarray option.

    .. versionadded:: 0.3.0

    Parameters
    ----------
    key: str
        The name of the option.

    Returns
    -------
    Any: the value of the option.
    """
    return OPTIONS[key]


class set_options:  # pylint: disable=invalid-name
    """
    Set the global rioxarray option.

    .. versionadded:: 0.3.0
    .. versionadded:: 0.7.0 skip_missing_spatial_dims
    .. versionadded:: 0.22.0 convention

    Parameters
    ----------
    export_grid_mapping: bool, default=True
        If True, this option will export the full Climate and Forecasts (CF)
        grid mapping attributes for the CRS. This is useful if you are exporting
        your file to netCDF using :meth:`xarray.Dataset.to_netcdf()`. When disabled,
        only the ``crs_wkt`` and ``spatial_ref`` attributes will be written and the
        program will be faster due to not needing to use
        :meth:`pyproj.CRS.to_cf() <pyproj.crs.CRS.to_cf>`.
    skip_missing_spatial_dims: bool, default=False
        If True, it will not perform spatial operations on variables
        within a :class:`xarray.Dataset` if the spatial dimensions
        are not found.
    convention: Convention, default=None
        The convention to use for reading and writing geospatial metadata.
        If None, CF convention is used as the default.
        See :class:`rioxarray.enum.Convention` for available options.


    Usage as a context manager::

        with rioxarray.set_options(export_grid_mapping=False):
            rds = rioxarray.open_rasterio(...)

    Usage for global settings::

        rioxarray.set_options(export_grid_mapping=False)

    """

    def __init__(self, **kwargs):
        self.old = OPTIONS.copy()
        for key, value in kwargs.items():
            if key not in OPTIONS:
                raise ValueError(
                    f"argument name {key} is not in the set of valid options "
                    f"{OPTION_NAMES}."
                )
            if key in VALIDATORS and not VALIDATORS[key](value):
                raise ValueError(f"option {key!r} gave an invalid value: {value!r}.")
            OPTIONS[key] = value

    def __enter__(self):
        return

    def __exit__(self, exc_type, exc_value, traceback):
        global OPTIONS  # pylint: disable=global-statement
        OPTIONS = self.old



"""Unit tests for the Zarr convention module."""
import numpy as np
import xarray as xr
from affine import Affine
from rasterio.crs import CRS

from rioxarray._convention import zarr
from rioxarray._convention.zarr import ZarrConvention


def test_has_convention_declared__proj():
    """Test checking for proj: convention declaration."""
    attrs = {
        "zarr_conventions": [
            {
                "name": "proj:",
                "uuid": "f17cb550-5864-4468-aeb7-f3180cfb622f",
            }
        ]
    }
    assert zarr.has_convention_declared(attrs, "proj:") is True
    assert zarr.has_convention_declared(attrs, "spatial:") is False


def test_has_convention_declared__spatial():
    """Test checking for spatial: convention declaration."""
    attrs = {
        "zarr_conventions": [
            {
                "name": "spatial:",
                "uuid": "689b58e2-cf7b-45e0-9fff-9cfc0883d6b4",
            }
        ]
    }
    assert zarr.has_convention_declared(attrs, "spatial:") is True
    assert zarr.has_convention_declared(attrs, "proj:") is False


def test_has_convention_declared__not_declared():
    """Test when no convention is declared."""
    attrs = {}
    assert zarr.has_convention_declared(attrs, "proj:") is False
    assert zarr.has_convention_declared(attrs, "spatial:") is False


def test_get_declared_conventions():
    """Test getting all declared conventions."""
    attrs = {
        "zarr_conventions": [
            {"name": "proj:", "uuid": "test-uuid-1"},
            {"name": "spatial:", "uuid": "test-uuid-2"},
        ]
    }
    declared = zarr.get_declared_conventions(attrs)
    assert declared == {"proj:", "spatial:"}


def test_parse_spatial_transform():
    """Test parsing spatial:transform array."""
    transform_array = [1.0, 0.0, 100.0, 0.0, -1.0, 200.0]
    result = zarr.parse_spatial_transform(transform_array)
    assert result == Affine(1.0, 0.0, 100.0, 0.0, -1.0, 200.0)


def test_parse_spatial_transform__invalid():
    """Test parsing invalid spatial:transform."""
    assert zarr.parse_spatial_transform([1, 2, 3]) is None
    assert zarr.parse_spatial_transform("invalid") is None


def test_read_crs__from_wkt2():
    """Test reading CRS from proj:wkt2 attribute."""
    data = xr.DataArray(np.random.rand(10, 10), dims=["y", "x"])
    data.attrs["zarr_conventions"] = [zarr.PROJ_CONVENTION]
    data.attrs["proj:wkt2"] = CRS.from_epsg(4326).to_wkt()

    crs = ZarrConvention.read_crs(data)
    assert crs is not None
    assert crs == CRS.from_epsg(4326)


def test_read_crs__from_code():
    """Test reading CRS from proj:code attribute."""
    data = xr.DataArray(np.random.rand(10, 10), dims=["y", "x"])
    data.attrs["zarr_conventions"] = [zarr.PROJ_CONVENTION]
    data.attrs["proj:code"] = "EPSG:4326"

    crs = ZarrConvention.read_crs(data)
    assert crs is not None
    assert crs == CRS.from_epsg(4326)


def test_read_crs__not_found():
    """Test that None is returned when no CRS is found."""
    data = xr.DataArray(np.random.rand(10, 10), dims=["y", "x"])

    crs = ZarrConvention.read_crs(data)
    assert crs is None


def test_read_crs__no_convention_declared():
    """Test that CRS is not read when convention is not declared."""
    data = xr.DataArray(np.random.rand(10, 10), dims=["y", "x"])
    # Add proj attributes but no convention declaration
    data.attrs["proj:wkt2"] = CRS.from_epsg(4326).to_wkt()

    crs = ZarrConvention.read_crs(data)
    assert crs is None


def test_read_transform__from_spatial_transform():
    """Test reading transform from spatial:transform attribute."""
    data = xr.DataArray(np.random.rand(10, 10), dims=["y", "x"])
    data.attrs["zarr_conventions"] = [zarr.SPATIAL_CONVENTION]
    data.attrs["spatial:transform"] = [1.0, 0.0, 100.0, 0.0, -1.0, 200.0]

    transform = ZarrConvention.read_transform(data)
    assert transform is not None
    assert transform == Affine(1.0, 0.0, 100.0, 0.0, -1.0, 200.0)


def test_read_transform__not_found():
    """Test that None is returned when no transform is found."""
    data = xr.DataArray(np.random.rand(10, 10), dims=["y", "x"])

    transform = ZarrConvention.read_transform(data)
    assert transform is None


def test_read_transform__no_convention_declared():
    """Test that transform is not read when convention is not declared."""
    data = xr.DataArray(np.random.rand(10, 10), dims=["y", "x"])
    # Add spatial attributes but no convention declaration
    data.attrs["spatial:transform"] = [1.0, 0.0, 100.0, 0.0, -1.0, 200.0]

    transform = ZarrConvention.read_transform(data)
    assert transform is None


def test_read_spatial_dimensions():
    """Test reading spatial dimensions from spatial:dimensions attribute."""
    data = xr.DataArray(np.random.rand(10, 20), dims=["lat", "lon"])
    data.attrs["zarr_conventions"] = [zarr.SPATIAL_CONVENTION]
    data.attrs["spatial:dimensions"] = ["lat", "lon"]

    dims = ZarrConvention.read_spatial_dimensions(data)
    assert dims == ("lat", "lon")


def test_read_spatial_dimensions__not_found():
    """Test that None is returned when no spatial dimensions are found."""
    data = xr.DataArray(np.random.rand(10, 10), dims=["y", "x"])

    dims = ZarrConvention.read_spatial_dimensions(data)
    assert dims is None


def test_read_spatial_dimensions__no_convention_declared():
    """Test that spatial dims are not read when convention is not declared."""
    data = xr.DataArray(np.random.rand(10, 10), dims=["y", "x"])
    # Add spatial attributes but no convention declaration
    data.attrs["spatial:dimensions"] = ["y", "x"]

    dims = ZarrConvention.read_spatial_dimensions(data)
    assert dims is None



import pytest

from rioxarray import set_options
from rioxarray._options import CONVENTION, EXPORT_GRID_MAPPING, get_option
from rioxarray.enum import Convention


def test_set_options__contextmanager():
    assert get_option(EXPORT_GRID_MAPPING)
    with set_options(**{EXPORT_GRID_MAPPING: False}):
        assert not get_option(EXPORT_GRID_MAPPING)
    assert get_option(EXPORT_GRID_MAPPING)


def test_set_options__global():
    assert get_option(EXPORT_GRID_MAPPING)
    try:
        set_options(export_grid_mapping=False)
        assert not get_option(EXPORT_GRID_MAPPING)
    finally:
        set_options(export_grid_mapping=True)
    assert get_option(EXPORT_GRID_MAPPING)


def test_set_options__invalid_argument():
    with pytest.raises(
        ValueError,
        match="argument name does_not_exist is not in the set of valid options",
    ):
        with set_options(does_not_exist=False):
            pass


def test_set_options__invalid_value():
    with pytest.raises(
        ValueError,
        match="option 'export_grid_mapping' gave an invalid value: 12345.",
    ):
        with set_options(export_grid_mapping=12345):
            pass


def test_set_options__convention_default():
    """Test that convention defaults to None."""
    assert get_option(CONVENTION) is None


def test_set_options__convention_cf():
    """Test setting convention to CF."""
    assert get_option(CONVENTION) is None
    with set_options(convention=Convention.CF):
        assert get_option(CONVENTION) is Convention.CF
    assert get_option(CONVENTION) is None


def test_set_options__convention_zarr():
    """Test setting convention to Zarr."""
    assert get_option(CONVENTION) is None
    with set_options(convention=Convention.ZARR):
        assert get_option(CONVENTION) is Convention.ZARR
    assert get_option(CONVENTION) is None


def test_set_options__convention_none():
    """Test setting convention back to None."""
    with set_options(convention=Convention.CF):
        assert get_option(CONVENTION) is Convention.CF
        with set_options(convention=None):
            assert get_option(CONVENTION) is None
        assert get_option(CONVENTION) is Convention.CF


def test_set_options__convention_invalid():
    """Test that invalid convention values raise error."""
    with pytest.raises(
        ValueError,
        match="option 'convention' gave an invalid value: 'invalid'.",
    ):
        with set_options(convention="invalid"):
            pass



"""
Zarr spatial and proj convention support for rioxarray.

This module provides functions for reading geospatial metadata according to:
- Zarr spatial convention: https://github.com/zarr-conventions/spatial
- Zarr geo-proj convention: https://github.com/zarr-conventions/geo-proj
"""
from typing import Optional, Union

import rasterio.crs
import xarray
from affine import Affine

from rioxarray.crs import crs_from_user_input

# Convention identifiers
PROJ_CONVENTION = {
    "schema_url": "https://raw.githubusercontent.com/zarr-conventions/geo-proj/refs/tags/v1/schema.json",
    "spec_url": "https://github.com/zarr-conventions/geo-proj/blob/v1/README.md",
    "uuid": "f17cb550-5864-4468-aeb7-f3180cfb622f",
    "name": "proj:",
    "description": "Coordinate reference system information for geospatial data",
}

SPATIAL_CONVENTION = {
    "schema_url": "https://raw.githubusercontent.com/zarr-conventions/spatial/refs/tags/v1/schema.json",
    "spec_url": "https://github.com/zarr-conventions/spatial/blob/v1/README.md",
    "uuid": "689b58e2-cf7b-45e0-9fff-9cfc0883d6b4",
    "name": "spatial:",
    "description": "Spatial coordinate information",
}


def has_convention_declared(attrs: dict, convention_name: str) -> bool:
    """
    Check if a specific convention is declared in zarr_conventions.

    Parameters
    ----------
    attrs : dict
        Attributes dictionary to check
    convention_name : str
        Name of convention to check for (e.g., "proj:" or "spatial:")

    Returns
    -------
    bool
        True if convention is declared
    """
    zarr_conventions = attrs.get("zarr_conventions", [])
    if not isinstance(zarr_conventions, list):
        return False

    for convention in zarr_conventions:
        if isinstance(convention, dict) and convention.get("name") == convention_name:
            return True
    return False


def get_declared_conventions(attrs: dict) -> set:
    """
    Get set of declared convention names from attrs.

    Parameters
    ----------
    attrs : dict
        Attributes dictionary to check

    Returns
    -------
    set
        Set of declared convention names (e.g., {"proj:", "spatial:"})
    """
    zarr_conventions = attrs.get("zarr_conventions", [])
    if not isinstance(zarr_conventions, list):
        return set()

    declared = set()
    for convention in zarr_conventions:
        if isinstance(convention, dict) and "name" in convention:
            declared.add(convention["name"])

    return declared


# ============================================================================
# Parsing utilities
# ============================================================================


def parse_spatial_transform(
    spatial_transform: Union[list, tuple],
) -> Optional[Affine]:
    """
    Convert spatial:transform array to Affine object.

    Parameters
    ----------
    spatial_transform : list or tuple
        Transform as [a, b, c, d, e, f] array

    Returns
    -------
    affine.Affine or None
        Affine transform object, or None if invalid
    """
    if not isinstance(spatial_transform, (list, tuple)):
        return None
    if len(spatial_transform) != 6:
        return None
    try:
        return Affine(*spatial_transform)
    except (TypeError, ValueError):
        return None


# ============================================================================
# Internal parsing helpers
# ============================================================================


def _parse_crs_from_attrs(
    attrs: dict, convention_check: bool = True
) -> Optional[rasterio.crs.CRS]:
    """
    Parse CRS from proj: attributes with fallback priority.

    Parameters
    ----------
    attrs : dict
        Attributes dictionary to parse from
    convention_check : bool, default True
        Whether to check for convention declaration

    Returns
    -------
    rasterio.crs.CRS or None
        Parsed CRS object, or None if not found
    """
    if convention_check and not has_convention_declared(attrs, "proj:"):
        return None

    # Try proj attributes in priority order: wkt2, code, projjson
    for proj_attr in ("proj:wkt2", "proj:code", "proj:projjson"):
        try:
            proj_value = attrs.get(proj_attr)
            if proj_value is not None:
                parsed_crs = crs_from_user_input(proj_value)
                if parsed_crs is not None:
                    return parsed_crs
        except (KeyError, TypeError, ValueError):
            pass
    return None


def _parse_transform_from_attrs(
    attrs: dict, convention_check: bool = True
) -> Optional[Affine]:
    """
    Parse transform from spatial: attributes.

    Parameters
    ----------
    attrs : dict
        Attributes dictionary to parse from
    convention_check : bool, default True
        Whether to check for convention declaration

    Returns
    -------
    affine.Affine or None
        Parsed transform object, or None if not found
    """
    if convention_check and not has_convention_declared(attrs, "spatial:"):
        return None

    try:
        spatial_transform = attrs.get("spatial:transform")
        if spatial_transform is not None:
            return parse_spatial_transform(spatial_transform)
    except (KeyError, TypeError, ValueError):
        pass
    return None


# ============================================================================
# ZarrConvention class implementing ConventionProtocol
# ============================================================================


class ZarrConvention:
    """Zarr convention class implementing ConventionProtocol."""

    @classmethod
    def read_crs(
        cls, obj: Union[xarray.Dataset, xarray.DataArray]
    ) -> Optional[rasterio.crs.CRS]:
        """
        Read CRS from Zarr proj: convention.

        Parameters
        ----------
        obj : xarray.Dataset or xarray.DataArray
            Object to read CRS from

        Returns
        -------
        rasterio.crs.CRS or None
            CRS object, or None if not found
        """
        return _parse_crs_from_attrs(obj.attrs)

    @classmethod
    def read_transform(
        cls, obj: Union[xarray.Dataset, xarray.DataArray]
    ) -> Optional[Affine]:
        """
        Read transform from Zarr spatial: convention.

        Parameters
        ----------
        obj : xarray.Dataset or xarray.DataArray
            Object to read transform from

        Returns
        -------
        affine.Affine or None
            Transform object, or None if not found
        """
        return _parse_transform_from_attrs(obj.attrs)

    @classmethod
    def read_spatial_dimensions(
        cls, obj: Union[xarray.Dataset, xarray.DataArray]
    ) -> Optional[tuple[str, str]]:
        """
        Read spatial dimensions from Zarr spatial: convention.

        Parameters
        ----------
        obj : xarray.Dataset or xarray.DataArray
            Object to read spatial dimensions from

        Returns
        -------
        tuple of (y_dim, x_dim) or None
            Tuple of dimension names, or None if not found
        """
        # Only interpret spatial:* attributes if convention is declared
        if not has_convention_declared(obj.attrs, "spatial:"):
            return None

        try:
            spatial_dims = obj.attrs.get("spatial:dimensions")
            if spatial_dims is not None and len(spatial_dims) >= 2:
                # spatial:dimensions format is ["y", "x"] or similar
                y_dim_name, x_dim_name = spatial_dims[-2:]  # Take last two
                if y_dim_name in obj.dims and x_dim_name in obj.dims:
                    return y_dim_name, x_dim_name
        except (KeyError, TypeError, ValueError):
            pass

        return None

    @classmethod
    def write_crs(
        cls,
        obj: Union[xarray.Dataset, xarray.DataArray],
        crs: rasterio.crs.CRS,
        **kwargs,  # pylint: disable=unused-argument
    ) -> Union[xarray.Dataset, xarray.DataArray]:
        """
        Write CRS using Zarr conventions.

        Note: Writing support will be implemented in a future PR.

        Parameters
        ----------
        obj : xarray.Dataset or xarray.DataArray
            Object to write CRS to
        crs : rasterio.crs.CRS
            CRS to write
        **kwargs
            Additional convention-specific parameters

        Returns
        -------
        xarray.Dataset or xarray.DataArray
            Object with CRS written

        Raises
        ------
        NotImplementedError
            Zarr write support is not yet implemented
        """
        raise NotImplementedError(
            "Zarr CRS writing is not yet implemented. "
            "Use Convention.CF for writing or wait for a future release."
        )

    @classmethod
    def write_transform(
        cls,
        obj: Union[xarray.Dataset, xarray.DataArray],
        *,
        transform: Affine,
        **kwargs,  # pylint: disable=unused-argument
    ) -> Union[xarray.Dataset, xarray.DataArray]:
        """
        Write transform using Zarr conventions.

        Note: Writing support will be implemented in a future PR.

        Parameters
        ----------
        obj : xarray.Dataset or xarray.DataArray
            Object to write transform to
        transform : affine.Affine
            Transform to write
        **kwargs
            Additional convention-specific parameters

        Returns
        -------
        xarray.Dataset or xarray.DataArray
            Object with transform written

        Raises
        ------
        NotImplementedError
            Zarr write support is not yet implemented
        """
        raise NotImplementedError(
            "Zarr transform writing is not yet implemented. "
            "Use Convention.CF for writing or wait for a future release."
        )



"""
CF (Climate and Forecasts) convention support for rioxarray.

This module provides functions for reading and writing geospatial metadata according to
the CF conventions: https://github.com/cf-convention/cf-conventions
"""
from typing import Optional, Tuple, Union

import numpy
import pyproj
import rasterio.crs
import xarray
from affine import Affine

from rioxarray._options import EXPORT_GRID_MAPPING, get_option
from rioxarray.crs import crs_from_user_input


class CFConvention:
    """CF convention class implementing ConventionProtocol."""

    @classmethod
    def read_crs(
        cls, obj: Union[xarray.Dataset, xarray.DataArray]
    ) -> Optional[rasterio.crs.CRS]:
        """
        Read CRS from CF conventions.

        Parameters
        ----------
        obj : xarray.Dataset or xarray.DataArray
            Object to read CRS from

        Returns
        -------
        rasterio.crs.CRS or None
            CRS object, or None if not found
        """
        try:
            grid_mapping_coord = obj.coords[obj.rio.grid_mapping]

            # Look in wkt attributes first for performance
            for crs_attr in ("spatial_ref", "crs_wkt"):
                try:
                    return crs_from_user_input(grid_mapping_coord.attrs[crs_attr])
                except KeyError:
                    pass

            # Look in grid_mapping CF attributes
            try:
                return pyproj.CRS.from_cf(grid_mapping_coord.attrs)
            except (KeyError, pyproj.exceptions.CRSError):
                pass
        except KeyError:
            # grid_mapping coordinate doesn't exist
            pass
        return None

    @classmethod
    def read_transform(
        cls, obj: Union[xarray.Dataset, xarray.DataArray]
    ) -> Optional[Affine]:
        """
        Read transform from CF conventions (GeoTransform attribute).

        Parameters
        ----------
        obj : xarray.Dataset or xarray.DataArray
            Object to read transform from

        Returns
        -------
        affine.Affine or None
            Transform object, or None if not found
        """
        try:
            transform = numpy.fromstring(
                obj.coords[obj.rio.grid_mapping].attrs["GeoTransform"], sep=" "
            )
            # Calling .tolist() to assure the arguments are Python float and JSON serializable
            return Affine.from_gdal(*transform.tolist())
        except KeyError:
            pass

        return None

    @classmethod
    def read_spatial_dimensions(
        cls,
        obj: Union[xarray.Dataset, xarray.DataArray],
    ) -> Optional[Tuple[str, str]]:
        """
        Read spatial dimensions from CF conventions.

        This function detects spatial dimensions based on:
        1. Standard dimension names ('x'/'y', 'longitude'/'latitude')
        2. CF coordinate attributes ('axis', 'standard_name')

        Parameters
        ----------
        obj : xarray.Dataset or xarray.DataArray
            Object to read spatial dimensions from

        Returns
        -------
        tuple of (y_dim, x_dim) or None
            Tuple of dimension names, or None if not found
        """
        x_dim = None
        y_dim = None

        # Check standard dimension names
        if "x" in obj.dims and "y" in obj.dims:
            return "y", "x"
        if "longitude" in obj.dims and "latitude" in obj.dims:
            return "latitude", "longitude"

        # Look for coordinates with CF attributes
        for coord in obj.coords:
            # Make sure to only look in 1D coordinates
            # that has the same dimension name as the coordinate
            if obj.coords[coord].dims != (coord,):
                continue
            if (obj.coords[coord].attrs.get("axis", "").upper() == "X") or (
                obj.coords[coord].attrs.get("standard_name", "").lower()
                in ("longitude", "projection_x_coordinate")
            ):
                x_dim = coord
            elif (obj.coords[coord].attrs.get("axis", "").upper() == "Y") or (
                obj.coords[coord].attrs.get("standard_name", "").lower()
                in ("latitude", "projection_y_coordinate")
            ):
                y_dim = coord

        if x_dim is not None and y_dim is not None:
            return str(y_dim), str(x_dim)
        return None

    @classmethod
    def write_crs(
        cls,
        obj: Union[xarray.Dataset, xarray.DataArray],
        crs: rasterio.crs.CRS,
        *,
        grid_mapping_name: Optional[str] = None,
        **kwargs,  # pylint: disable=unused-argument
    ) -> Union[xarray.Dataset, xarray.DataArray]:
        """
        Write CRS using CF conventions.

        This also writes the grid_mapping attribute to encoding for CF compliance.

        Parameters
        ----------
        obj : xarray.Dataset or xarray.DataArray
            Object to write CRS to
        crs : rasterio.crs.CRS
            CRS to write
        grid_mapping_name : str
            Name of the grid_mapping coordinate
        **kwargs
            Additional convention-specific parameters

        Returns
        -------
        xarray.Dataset or xarray.DataArray
            Object with CRS written
        """
        # get original transform
        transform = obj.rio._cached_transform()
        # remove old grid mapping coordinate if exists
        grid_mapping_name = (
            obj.rio.grid_mapping if grid_mapping_name is None else grid_mapping_name
        )
        try:
            del obj.coords[grid_mapping_name]
        except KeyError:
            pass

        # add grid mapping coordinate
        obj.coords[grid_mapping_name] = xarray.Variable((), 0)
        grid_map_attrs = {}
        if get_option(EXPORT_GRID_MAPPING):
            try:
                grid_map_attrs = pyproj.CRS.from_user_input(crs).to_cf()
            except KeyError:
                pass
        # spatial_ref is for compatibility with GDAL
        crs_wkt = crs.to_wkt()
        grid_map_attrs["spatial_ref"] = crs_wkt
        grid_map_attrs["crs_wkt"] = crs_wkt
        if transform is not None:
            grid_map_attrs["GeoTransform"] = " ".join(
                [str(item) for item in transform.to_gdal()]
            )
        obj.coords[grid_mapping_name].rio.set_attrs(grid_map_attrs, inplace=True)

        return obj.rio.write_grid_mapping(
            grid_mapping_name=grid_mapping_name, inplace=True
        )

    @classmethod
    def write_transform(
        cls,
        obj: Union[xarray.Dataset, xarray.DataArray],
        *,
        transform: Affine,
        grid_mapping_name: Optional[str] = None,
        **kwargs,  # pylint: disable=unused-argument
    ) -> Union[xarray.Dataset, xarray.DataArray]:
        """
        Write transform using CF conventions (GeoTransform attribute).

        This also writes the grid_mapping attribute to encoding for CF compliance.

        Parameters
        ----------
        obj : xarray.Dataset or xarray.DataArray
            Object to write transform to
        transform : affine.Affine
            Transform to write
        grid_mapping_name : Optional[str]
            Name of the grid_mapping coordinate
        **kwargs
            Additional convention-specific parameters

        Returns
        -------
        xarray.Dataset or xarray.DataArray
            Object with transform written
        """
        transform = transform or obj.rio.transform(recalc=True)
        grid_mapping_name = (
            obj.rio.grid_mapping if grid_mapping_name is None else grid_mapping_name
        )
        try:
            grid_map_attrs = obj.coords[grid_mapping_name].attrs.copy()
        except KeyError:
            obj.coords[grid_mapping_name] = xarray.Variable((), 0)
            grid_map_attrs = obj.coords[grid_mapping_name].attrs.copy()
        grid_map_attrs["GeoTransform"] = " ".join(
            [str(item) for item in transform.to_gdal()]
        )
        obj.coords[grid_mapping_name].rio.set_attrs(grid_map_attrs, inplace=True)
        return obj.rio.write_grid_mapping(
            grid_mapping_name=grid_mapping_name, inplace=True
        )



"""
This module is an extension for xarray to provide rasterio capabilities
to xarray datasets/dataarrays.
"""

# pylint: disable=too-many-lines
import json
import math
import warnings
from collections.abc import Hashable, Iterable
from typing import Any, Literal, Optional, Union

import numpy
import rasterio.warp
import rasterio.windows
import xarray
from affine import Affine
from pyproj.aoi import AreaOfInterest
from pyproj.database import query_utm_crs_info
from rasterio.control import GroundControlPoint
from rasterio.crs import CRS
from rasterio.rpc import RPC

from rioxarray._convention._core import (
    _get_convention,
    read_crs_auto,
    read_spatial_dimensions_auto,
    read_transform_auto,
)
from rioxarray._spatial_utils import (  # noqa: F401, pylint: disable=unused-import
    DEFAULT_GRID_MAP,
    _affine_has_rotation,
    _convert_gcps_to_geojson,
    _get_data_var_message,
    _get_spatial_dims,
    _has_spatial_dims,
    _order_bounds,
    _resolution,
    affine_to_coords,
)
from rioxarray.crs import crs_from_user_input
from rioxarray.enum import Convention
from rioxarray.exceptions import (
    DimensionError,
    DimensionMissingCoordinateError,
    InvalidDimensionOrder,
    MissingCRS,
    MissingSpatialDimensionError,
    NoDataInBounds,
    OneDimensionalRaster,
    RioXarrayError,
    TooManyDimensions,
)


class XRasterBase:
    """This is the base class for the GIS extensions for xarray"""

    # pylint: disable=too-many-instance-attributes
    def __init__(self, xarray_obj: Union[xarray.DataArray, xarray.Dataset]):
        self._obj: Union[xarray.DataArray, xarray.Dataset] = xarray_obj

        self._x_dim: Optional[Hashable] = None
        self._y_dim: Optional[Hashable] = None

        # Auto-detect spatial dimensions from any supported convention
        spatial_dims = read_spatial_dimensions_auto(self._obj)
        if spatial_dims is not None:
            self._y_dim, self._x_dim = spatial_dims

        # properties
        self._count: Optional[int] = None
        self._height: Optional[int] = None
        self._width: Optional[int] = None
        self._crs: Union[rasterio.crs.CRS, None, Literal[False]] = None
        self._gcps: Optional[list[GroundControlPoint]] = None
        self._rpcs: Optional[RPC] = None

    @property
    def crs(self) -> Optional[rasterio.crs.CRS]:
        """:obj:`rasterio.crs.CRS`:
        Retrieve projection from :obj:`xarray.Dataset` | :obj:`xarray.DataArray`
        """
        if self._crs is not None:
            return None if self._crs is False else self._crs

        # Auto-detect CRS from any supported convention
        parsed_crs = read_crs_auto(self._obj)

        if parsed_crs is not None:
            self._set_crs(parsed_crs, inplace=True)
            return self._crs

        # No CRS found
        self._crs = False
        return None

    def _get_obj(self, inplace: bool) -> Union[xarray.Dataset, xarray.DataArray]:
        """
        Get the object to modify.

        Parameters
        ----------
        inplace: bool
            If True, returns self.

        Returns
        -------
        :obj:`xarray.Dataset` | :obj:`xarray.DataArray`
        """
        if inplace:
            return self._obj
        obj_copy = self._obj.copy(deep=True)
        # preserve attribute information
        obj_copy.rio._x_dim = self._x_dim
        obj_copy.rio._y_dim = self._y_dim
        obj_copy.rio._width = self._width
        obj_copy.rio._height = self._height
        obj_copy.rio._crs = self._crs
        obj_copy.rio._gcps = self._gcps
        obj_copy.rio._rpcs = self._rpcs
        return obj_copy

    def set_crs(
        self, input_crs: Any, inplace: bool = True
    ) -> Union[xarray.Dataset, xarray.DataArray]:
        """
        Set the CRS value for the Dataset/DataArray without modifying
        the dataset/data array.

        .. deprecated:: 0.15.8
            It is recommended to use `rio.write_crs()` instead. This
        method will likely be removed in a future release.

        Parameters
        ----------
        input_crs: object
            Anything accepted by `rasterio.crs.CRS.from_user_input`.
        inplace: bool, optional
            If True, it will write to the existing dataset. Default is False.

        Returns
        -------
        :obj:`xarray.Dataset` | :obj:`xarray.DataArray`:
            Dataset with crs attribute.
        """
        warnings.warn(
            "It is recommended to use 'rio.write_crs()' instead. 'rio.set_crs()' will likely"
            "be removed in a future release.",
            FutureWarning,
            stacklevel=2,
        )

        return self._set_crs(input_crs, inplace=inplace)

    def _set_crs(
        self, input_crs: Any, inplace: bool = True
    ) -> Union[xarray.Dataset, xarray.DataArray]:
        """
        Set the CRS value for the Dataset/DataArray without modifying
        the dataset/data array.

        Parameters
        ----------
        input_crs: object
            Anything accepted by `rasterio.crs.CRS.from_user_input`.
        inplace: bool, optional
            If True, it will write to the existing dataset. Default is False.

        Returns
        -------
        xarray.Dataset | xarray.DataArray
            Dataset with crs attribute.
        """
        crs = crs_from_user_input(input_crs)
        obj = self._get_obj(inplace=inplace)
        obj.rio._crs = crs
        return obj

    @property
    def grid_mapping(self) -> str:
        """
        str: The CF grid_mapping attribute. 'spatial_ref' is the default.
        """
        grid_mapping = self._obj.encoding.get(
            "grid_mapping", self._obj.attrs.get("grid_mapping")
        )
        if grid_mapping is not None:
            return grid_mapping
        grid_mapping = DEFAULT_GRID_MAP
        # search the dataset for the grid mapping name
        if hasattr(self._obj, "data_vars"):
            grid_mappings = set()
            for var in self._obj.data_vars:
                if not _has_spatial_dims(self._obj, var=var):
                    continue
                var_grid_mapping = self._obj[var].encoding.get(
                    "grid_mapping", self._obj[var].attrs.get("grid_mapping")
                )
                if var_grid_mapping is not None:
                    grid_mapping = var_grid_mapping
                    grid_mappings.add(grid_mapping)
            if len(grid_mappings) > 1:
                raise RioXarrayError("Multiple grid mappings exist.")
        return grid_mapping

    def write_grid_mapping(
        self,
        grid_mapping_name: str = DEFAULT_GRID_MAP,
        inplace: bool = False,
    ) -> Union[xarray.Dataset, xarray.DataArray]:
        """
        Write the grid_mapping attribute to the encoding.

        Parameters
        ----------
        grid_mapping_name: str
            Name of the grid_mapping coordinate.
        inplace: bool, optional
            If True, it will write to the existing dataset. Default is False.

        Returns
        -------
        :obj:`xarray.Dataset` | :obj:`xarray.DataArray`:
            Modified dataset with grid_mapping written.

        See Also
        --------
        :meth:`rioxarray.rioxarray.XRasterBase.write_crs`
        """
        data_obj = self._get_obj(inplace=inplace)
        if hasattr(data_obj, "data_vars"):
            for var in data_obj.data_vars:
                try:
                    x_dim, y_dim = _get_spatial_dims(data_obj, var=var)
                except MissingSpatialDimensionError:
                    continue
                # remove grid_mapping from attributes if it exists
                # and update the grid_mapping in encoding
                new_attrs = dict(data_obj[var].attrs)
                new_attrs.pop("grid_mapping", None)
                data_obj[var].rio.update_encoding(
                    {"grid_mapping": grid_mapping_name}, inplace=True
                ).rio.set_attrs(new_attrs, inplace=True).rio.set_spatial_dims(
                    x_dim=x_dim, y_dim=y_dim, inplace=True
                )
        # remove grid_mapping from attributes if it exists
        # and update the grid_mapping in encoding
        new_attrs = dict(data_obj.attrs)
        new_attrs.pop("grid_mapping", None)
        return data_obj.rio.update_encoding(
            {"grid_mapping": grid_mapping_name}, inplace=True
        ).rio.set_attrs(new_attrs, inplace=True)

    def write_crs(
        self,
        input_crs: Optional[Any] = None,
        grid_mapping_name: Optional[str] = None,
        convention: Optional[Convention] = None,
        inplace: bool = False,
    ) -> Union[xarray.Dataset, xarray.DataArray]:
        """
        Write the CRS to the dataset using the specified convention.

        .. warning:: When using CF convention, the grid_mapping attribute is written to the encoding.

        Parameters
        ----------
        input_crs: Any
            Anything accepted by `rasterio.crs.CRS.from_user_input`.
        grid_mapping_name: str, optional
            Name of the grid_mapping coordinate to store the CRS information in.
            Only used with CF convention. Default is the grid_mapping name of the dataset.
        convention: Convention, optional
            Convention to use for writing CRS. If None, uses the global default
            from set_options(). Currently only CF convention is supported.
        inplace: bool, optional
            If True, it will write to the existing dataset. Default is False.

        Returns
        -------
        :obj:`xarray.Dataset` | :obj:`xarray.DataArray`:
            Modified dataset with CRS information.

        Examples
        --------
        Write the CRS of the current `xarray` object:

        >>> raster.rio.write_crs("epsg:4326", inplace=True)

        Write the CRS on a copy:

        >>> raster = raster.rio.write_crs("epsg:4326")
        """
        if input_crs is not None:
            data_obj = self._set_crs(input_crs, inplace=inplace)
        else:
            data_obj = self._get_obj(inplace=inplace)

        if data_obj.rio.crs is None:
            raise MissingCRS(
                "CRS not found. Please set the CRS with 'rio.write_crs()'."
            )

        # Remove legacy crs attr (not part of any convention)
        data_obj.attrs.pop("crs", None)

        # Use the convention module to write CRS
        # Pass user input grid_mapping_name (may be None, convention handles default)
        convention_module = _get_convention(convention)
        return convention_module.write_crs(
            data_obj,
            crs=data_obj.rio.crs,
            grid_mapping_name=grid_mapping_name,
        )

    def estimate_utm_crs(self, datum_name: str = "WGS 84") -> rasterio.crs.CRS:
        """Returns the estimated UTM CRS based on the bounds of the dataset.

        .. versionadded:: 0.2

        .. note:: Requires pyproj 3+

        Parameters
        ----------
        datum_name : str, optional
            The name of the datum to use in the query. Default is WGS 84.

        Returns
        -------
        rasterio.crs.CRS
        """
        if self.crs is None:
            raise RuntimeError("crs must be set to estimate UTM CRS.")

        # ensure using geographic coordinates
        if self.crs.is_geographic:  # pylint: disable=no-member
            minx, miny, maxx, maxy = self.bounds(recalc=True)
        else:
            minx, miny, maxx, maxy = self.transform_bounds("EPSG:4326", recalc=True)

        x_center = numpy.mean([minx, maxx]).item()
        y_center = numpy.mean([miny, maxy]).item()

        utm_crs_list = query_utm_crs_info(
            datum_name=datum_name,
            area_of_interest=AreaOfInterest(
                west_lon_degree=x_center,
                south_lat_degree=y_center,
                east_lon_degree=x_center,
                north_lat_degree=y_center,
            ),
        )
        try:
            return CRS.from_epsg(utm_crs_list[0].code)
        except IndexError:
            raise RuntimeError("Unable to determine UTM CRS") from None

    def _cached_transform(self) -> Optional[Affine]:
        """
        Get the transform by auto-detecting from any supported convention.
        """
        return read_transform_auto(self._obj)

    def write_transform(
        self,
        transform: Optional[Affine] = None,
        grid_mapping_name: Optional[str] = None,
        convention: Optional[Convention] = None,
        inplace: bool = False,
    ) -> Union[xarray.Dataset, xarray.DataArray]:
        """
        .. versionadded:: 0.0.30

        Write the transform to the dataset using the specified convention.

        For CF convention, this writes the GeoTransform to the dataset where GDAL can read it in.
        https://gdal.org/drivers/raster/netcdf.html#georeference

        Parameters
        ----------
        transform: affine.Affine, optional
            The transform of the dataset. If not provided, it will be calculated.
        grid_mapping_name: str, optional
            Name of the grid_mapping coordinate to store the transform information in.
            Only used with CF convention. Default is the grid_mapping name of the dataset.
        convention: Convention, optional
            Convention to use for writing transform. If None, uses the global default
            from set_options(). Currently only CF convention is supported.
        inplace: bool, optional
            If True, it will write to the existing dataset. Default is False.

        Returns
        -------
        :obj:`xarray.Dataset` | :obj:`xarray.DataArray`:
            Modified dataset with transform written.
        """
        transform = transform or self.transform(recalc=True)
        data_obj = self._get_obj(inplace=inplace)

        # Remove legacy transform attr (not part of any convention)
        data_obj.attrs.pop("transform", None)

        # Use the convention module to write transform
        # Pass user input grid_mapping_name (may be None, convention handles default)
        convention_module = _get_convention(convention)
        return convention_module.write_transform(
            data_obj,
            transform=transform,
            grid_mapping_name=grid_mapping_name,
        )

    def transform(self, recalc: bool = False) -> Affine:
        """
        Parameters
        ----------
        recalc: bool, optional
            If True, it will re-calculate the transform instead of using
            the cached transform.

        Returns
        -------
        :obj:`affine.Affine`:
            The affine of the :obj:`xarray.Dataset` | :obj:`xarray.DataArray`
        """
        transform = self._cached_transform()
        if transform and (
            not transform.is_rectilinear or _affine_has_rotation(transform)
        ):
            if recalc:
                warnings.warn(
                    "Transform that is non-rectilinear or with rotation found. "
                    "Unable to recalculate."
                )
            return transform

        try:
            src_left, _, _, src_top = self._unordered_bounds(recalc=recalc)
            src_resolution_x, src_resolution_y = self.resolution(recalc=recalc)
        except (DimensionMissingCoordinateError, DimensionError):
            return Affine.identity() if transform is None else transform
        return Affine.translation(src_left, src_top) * Affine.scale(
            src_resolution_x, src_resolution_y
        )

    def write_coordinate_system(
        self, inplace: bool = False
    ) -> Union[xarray.Dataset, xarray.DataArray]:
        """
        Write the coordinate system CF metadata.

        .. versionadded:: 0.0.30

        Parameters
        ----------
        inplace: bool, optional
            If True, it will write to the existing dataset. Default is False.

        Returns
        -------
        :obj:`xarray.Dataset` | :obj:`xarray.DataArray`:
            The dataset with the CF coordinate system attributes added.
        """
        data_obj = self._get_obj(inplace=inplace)
        # add metadata to x,y coordinates
        is_projected = data_obj.rio.crs and data_obj.rio.crs.is_projected
        is_geographic = data_obj.rio.crs and data_obj.rio.crs.is_geographic
        x_coord_attrs = dict(data_obj.coords[self.x_dim].attrs)
        x_coord_attrs["axis"] = "X"
        y_coord_attrs = dict(data_obj.coords[self.y_dim].attrs)
        y_coord_attrs["axis"] = "Y"
        if is_projected:
            units = None
            if hasattr(data_obj.rio.crs, "linear_units_factor"):
                unit_factor = data_obj.rio.crs.linear_units_factor[-1]
                if unit_factor != 1:
                    units = f"{unit_factor} metre"
                else:
                    units = "metre"
            # X metadata
            x_coord_attrs["long_name"] = "x coordinate of projection"
            x_coord_attrs["standard_name"] = "projection_x_coordinate"
            if units:
                x_coord_attrs["units"] = units
            # Y metadata
            y_coord_attrs["long_name"] = "y coordinate of projection"
            y_coord_attrs["standard_name"] = "projection_y_coordinate"
            if units:
                y_coord_attrs["units"] = units
        elif is_geographic:
            # X metadata
            x_coord_attrs["long_name"] = "longitude"
            x_coord_attrs["standard_name"] = "longitude"
            x_coord_attrs["units"] = "degrees_east"
            # Y metadata
            y_coord_attrs["long_name"] = "latitude"
            y_coord_attrs["standard_name"] = "latitude"
            y_coord_attrs["units"] = "degrees_north"
        data_obj.coords[self.y_dim].attrs = y_coord_attrs
        data_obj.coords[self.x_dim].attrs = x_coord_attrs
        return data_obj

    def set_attrs(
        self, new_attrs: dict, inplace: bool = False
    ) -> Union[xarray.Dataset, xarray.DataArray]:
        """
        Set the attributes of the dataset/dataarray and reset
        rioxarray properties to re-search for them.

        Parameters
        ----------
        new_attrs: dict
            A dictionary of new attributes.
        inplace: bool, optional
            If True, it will write to the existing dataset. Default is False.

        Returns
        -------
        :obj:`xarray.Dataset` | :obj:`xarray.DataArray`:
            Modified dataset with new attributes.
        """
        data_obj = self._get_obj(inplace=inplace)
        # set the attributes
        data_obj.attrs = new_attrs
        # reset rioxarray properties depending
        # on attributes to be generated
        data_obj.rio._nodata = None
        data_obj.rio._crs = None
        return data_obj

    def update_attrs(
        self, new_attrs: dict, inplace: bool = False
    ) -> Union[xarray.Dataset, xarray.DataArray]:
        """
        Update the attributes of the dataset/dataarray and reset
        rioxarray properties to re-search for them.

        Parameters
        ----------
        new_attrs: dict
            A dictionary of new attributes to update with.
        inplace: bool, optional
            If True, it will write to the existing dataset. Default is False.

        Returns
        -------
        :obj:`xarray.Dataset` | :obj:`xarray.DataArray`:
            Modified dataset with updated attributes.
        """
        data_attrs = dict(self._obj.attrs)
        data_attrs.update(**new_attrs)
        return self.set_attrs(data_attrs, inplace=inplace)

    def set_encoding(
        self, new_encoding: dict, inplace: bool = False
    ) -> Union[xarray.Dataset, xarray.DataArray]:
        """
        Set the encoding of the dataset/dataarray and reset
        rioxarray properties to re-search for them.

        .. versionadded:: 0.4

        Parameters
        ----------
        new_encoding: dict
            A dictionary for encoding.
        inplace: bool, optional
            If True, it will write to the existing dataset. Default is False.

        Returns
        -------
        :obj:`xarray.Dataset` | :obj:`xarray.DataArray`:
            Modified dataset with new attributes.
        """
        data_obj = self._get_obj(inplace=inplace)
        # set the attributes
        data_obj.encoding = new_encoding
        # reset rioxarray properties depending
        # on attributes to be generated
        data_obj.rio._nodata = None
        data_obj.rio._crs = None
        return data_obj

    def update_encoding(
        self, new_encoding: dict, inplace: bool = False
    ) -> Union[xarray.Dataset, xarray.DataArray]:
        """
        Update the encoding of the dataset/dataarray and reset
        rioxarray properties to re-search for them.

        .. versionadded:: 0.4

        Parameters
        ----------
        new_encoding: dict
            A dictionary with encoding values to update with.
        inplace: bool, optional
            If True, it will write to the existing dataset. Default is False.

        Returns
        -------
        :obj:`xarray.Dataset` | :obj:`xarray.DataArray`:
            Modified dataset with updated attributes.
        """
        data_encoding = dict(self._obj.encoding)
        data_encoding.update(**new_encoding)
        return self.set_encoding(data_encoding, inplace=inplace)

    def set_spatial_dims(
        self, x_dim: str, y_dim: str, inplace: bool = True
    ) -> Union[xarray.Dataset, xarray.DataArray]:
        """
        This sets the spatial dimensions of the dataset.

        Parameters
        ----------
        x_dim: str
            The name of the x dimension.
        y_dim: str
            The name of the y dimension.
        inplace: bool, optional
            If True, it will modify the dataframe in place.
            Otherwise it will return a modified copy.

        Returns
        -------
        :obj:`xarray.Dataset` | :obj:`xarray.DataArray`:
            Dataset with spatial dimensions set.
        """

        data_obj = self._get_obj(inplace=inplace)
        if x_dim in data_obj.dims:
            data_obj.rio._x_dim = x_dim
        else:
            raise MissingSpatialDimensionError(
                f"x dimension ({x_dim}) not found.{_get_data_var_message(data_obj)}"
            )
        if y_dim in data_obj.dims:
            data_obj.rio._y_dim = y_dim
        else:
            raise MissingSpatialDimensionError(
                f"y dimension ({y_dim}) not found.{_get_data_var_message(data_obj)}"
            )
        return data_obj

    @property
    def x_dim(self) -> Hashable:
        """Hashable: The dimension for the X-axis."""
        if self._x_dim is not None:
            return self._x_dim
        raise MissingSpatialDimensionError(
            "x dimension not found. 'rio.set_spatial_dims()' or "
            "using 'rename()' to change the dimension name to 'x' can address this."
            f"{_get_data_var_message(self._obj)}"
        )

    @property
    def y_dim(self) -> Hashable:
        """Hashable: The dimension for the Y-axis."""
        if self._y_dim is not None:
            return self._y_dim
        raise MissingSpatialDimensionError(
            "y dimension not found. 'rio.set_spatial_dims()' or "
            "using 'rename()' to change the dimension name to 'y' can address this."
            f"{_get_data_var_message(self._obj)}"
        )

    @property
    def width(self) -> int:
        """int: Returns the width of the dataset (x dimension size)"""
        if self._width is not None:
            return self._width
        self._width = self._obj[self.x_dim].size
        return self._width

    @property
    def height(self) -> int:
        """int: Returns the height of the dataset (y dimension size)"""
        if self._height is not None:
            return self._height
        self._height = self._obj[self.y_dim].size
        return self._height

    @property
    def shape(self) -> tuple[int, int]:
        """tuple(int, int): Returns the shape (height, width)"""
        return (self.height, self.width)

    def _check_dimensions(self) -> Optional[str]:
        """
        This function validates that the dimensions 2D/3D and
        they are are in the proper order.

        Returns
        -------
        str or None: Name extra dimension.
        """
        extra_dims = tuple(set(list(self._obj.dims)) - {self.x_dim, self.y_dim})
        if len(extra_dims) > 1:
            raise TooManyDimensions(
                "Only 2D and 3D data arrays supported."
                f"{_get_data_var_message(self._obj)}"
            )
        if extra_dims and self._obj.dims != (extra_dims[0], self.y_dim, self.x_dim):
            dim_info: tuple = (extra_dims[0], self.y_dim, self.x_dim)
            raise InvalidDimensionOrder(
                f"Invalid dimension order. Expected order: {dim_info}. "
                f"You can use `DataArray.transpose{dim_info}`"
                " to reorder your dimensions."
                f"{_get_data_var_message(self._obj)}"
            )
        if not extra_dims and self._obj.dims != (self.y_dim, self.x_dim):
            dim_info = (self.y_dim, self.x_dim)
            raise InvalidDimensionOrder(
                f"Invalid dimension order. Expected order: {dim_info}. "
                f"You can use `DataArray.transpose{dim_info}`"
                " to reorder your dimensions."
                f"{_get_data_var_message(self._obj)}"
            )
        return str(extra_dims[0]) if extra_dims else None

    @property
    def count(self) -> int:
        """int: Returns the band count (z dimension size)"""
        if self._count is not None:
            return self._count
        extra_dim = self._check_dimensions()
        self._count = 1
        if extra_dim is not None:
            self._count = self._obj[extra_dim].size
        return self._count

    def _internal_bounds(self) -> tuple[float, float, float, float]:
        """Determine the internal bounds of the `xarray.DataArray`"""
        if self.x_dim not in self._obj.coords:
            raise DimensionMissingCoordinateError(f"{self.x_dim} missing coordinates.")
        if self.y_dim not in self._obj.coords:
            raise DimensionMissingCoordinateError(f"{self.y_dim} missing coordinates.")
        try:
            left = float(self._obj[self.x_dim][0])
            right = float(self._obj[self.x_dim][-1])
            top = float(self._obj[self.y_dim][0])
            bottom = float(self._obj[self.y_dim][-1])
        except IndexError:
            raise NoDataInBounds(
                "Unable to determine bounds from coordinates."
                f"{_get_data_var_message(self._obj)}"
            ) from None
        return left, bottom, right, top

    def resolution(self, recalc: bool = False) -> tuple[float, float]:
        """
        Determine if the resolution of the grid.
        If the transformation has rotation, the sign of the resolution is lost.

        Parameters
        ----------
        recalc: bool, optional
            Will force the resolution to be recalculated instead of using the
            transform attribute.

        Returns
        -------
        x_resolution, y_resolution: float
            The resolution of the `xarray.DataArray` | `xarray.Dataset`
        """
        transform = self._cached_transform()

        if (
            not recalc or self.width == 1 or self.height == 1
        ) and transform is not None:
            return _resolution(transform)

        # if the coordinates of the spatial dimensions are missing
        # use the cached transform resolution
        try:
            left, bottom, right, top = self._internal_bounds()
        except DimensionMissingCoordinateError:
            if transform is None:
                raise
            return _resolution(transform)

        if self.width == 1 or self.height == 1:
            raise OneDimensionalRaster(
                "Only 1 dimenional array found. Cannot calculate the resolution."
                f"{_get_data_var_message(self._obj)}"
            )

        resolution_x = (right - left) / (self.width - 1)
        resolution_y = (bottom - top) / (self.height - 1)
        return resolution_x, resolution_y

    def _unordered_bounds(
        self, recalc: bool = False
    ) -> tuple[float, float, float, float]:
        """
        Unordered bounds.

        Parameters
        ----------
        recalc: bool, optional
            Will force the bounds to be recalculated instead of using the
            transform attribute.

        Returns
        -------
        left, bottom, right, top: float
            Outermost coordinates of the `xarray.DataArray` | `xarray.Dataset`.
        """
        resolution_x, resolution_y = self.resolution(recalc=recalc)

        try:
            # attempt to get bounds from xarray coordinate values
            left, bottom, right, top = self._internal_bounds()
            left -= resolution_x / 2.0
            right += resolution_x / 2.0
            top -= resolution_y / 2.0
            bottom += resolution_y / 2.0
        except DimensionMissingCoordinateError as error:
            transform = self._cached_transform()
            if not transform:
                raise RioXarrayError("Transform not able to be determined.") from error
            left = transform.c
            top = transform.f
            right = left + resolution_x * self.width
            bottom = top + resolution_y * self.height

        return left, bottom, right, top

    def bounds(self, *, recalc: bool = False) -> tuple[float, float, float, float]:
        """
        Parameters
        ----------
        recalc: bool, optional
            Will force the bounds to be recalculated instead of using the
            transform attribute.

        Returns
        -------
        left, bottom, right, top: float
            Outermost coordinates of the `xarray.DataArray` | `xarray.Dataset`.
        """
        minx, miny, maxx, maxy = self._unordered_bounds(recalc=recalc)
        resolution_x, resolution_y = self.resolution(recalc=recalc)
        return _order_bounds(
            minx=minx,
            miny=miny,
            maxx=maxx,
            maxy=maxy,
            resolution_x=resolution_x,
            resolution_y=resolution_y,
        )

    def isel_window(
        self, window: rasterio.windows.Window, *, pad: bool = False
    ) -> Union[xarray.Dataset, xarray.DataArray]:
        """
        Use a rasterio.windows.Window to select a subset of the data.

        .. versionadded:: 0.6.0 pad

        .. warning:: Float indices are converted to integers.

        Parameters
        ----------
        window: :class:`rasterio.windows.Window`
            The window of the dataset to read.
        pad: bool, default=False
            Set to True to expand returned DataArray to dimensions of the window

        Returns
        -------
        :obj:`xarray.Dataset` | :obj:`xarray.DataArray`:
            The data in the window.
        """
        (row_start, row_stop), (col_start, col_stop) = window.toranges()
        row_start = 0 if row_start < 0 else math.floor(row_start)
        row_stop = 0 if row_stop < 0 else math.ceil(row_stop)
        col_start = 0 if col_start < 0 else math.floor(col_start)
        col_stop = 0 if col_stop < 0 else math.ceil(col_stop)
        row_slice = slice(int(row_start), int(row_stop))
        col_slice = slice(int(col_start), int(col_stop))
        array_subset = (
            self._obj.isel({self.y_dim: row_slice, self.x_dim: col_slice})
            .copy()  # this is to prevent sharing coordinates with the original dataset
            .rio.set_spatial_dims(x_dim=self.x_dim, y_dim=self.y_dim, inplace=True)
            .rio.write_transform(
                transform=rasterio.windows.transform(
                    rasterio.windows.Window.from_slices(
                        rows=row_slice,
                        cols=col_slice,
                        width=self.width,
                        height=self.height,
                    ),
                    self.transform(recalc=True),
                ),
                inplace=True,
            )
        )
        if pad:
            return array_subset.rio.pad_box(
                *rasterio.windows.bounds(window, self.transform(recalc=True))
            )
        return array_subset

    def slice_xy(
        self,
        minx: float,
        miny: float,
        maxx: float,
        maxy: float,
    ) -> Union[xarray.Dataset, xarray.DataArray]:
        """Slice the array by x,y bounds.

        Parameters
        ----------
        minx: float
            Minimum bound for x coordinate.
        miny: float
            Minimum bound for y coordinate.
        maxx: float
            Maximum bound for x coordinate.
        maxy: float
            Maximum bound for y coordinate.


        Returns
        -------
        :obj:`xarray.Dataset` | :obj:`xarray.DataArray`:
            The data in the slice.
        """
        left, bottom, right, top = self._internal_bounds()
        if top > bottom:
            y_slice = slice(maxy, miny)
        else:
            y_slice = slice(miny, maxy)

        if left > right:
            x_slice = slice(maxx, minx)
        else:
            x_slice = slice(minx, maxx)

        subset = (
            self._obj.sel({self.x_dim: x_slice, self.y_dim: y_slice})
            .copy()  # this is to prevent sharing coordinates with the original dataset
            .rio.set_spatial_dims(x_dim=self.x_dim, y_dim=self.y_dim, inplace=True)
            .rio.write_transform(inplace=True)
        )
        return subset

    def transform_bounds(
        self, dst_crs: Any, *, densify_pts: int = 21, recalc: bool = False
    ) -> tuple[float, float, float, float]:
        """Transform bounds from src_crs to dst_crs.

        Optionally densifying the edges (to account for nonlinear transformations
        along these edges) and extracting the outermost bounds.

        Note: this does not account for the antimeridian.

        Parameters
        ----------
        dst_crs: str, :obj:`rasterio.crs.CRS`, or dict
            Target coordinate reference system.
        densify_pts: uint, optional
            Number of points to add to each edge to account for nonlinear
            edges produced by the transform process.  Large numbers will produce
            worse performance.  Default: 21 (gdal default).
        recalc: bool, optional
            Will force the bounds to be recalculated instead of using the transform
            attribute.

        Returns
        -------
        left, bottom, right, top: float
            Outermost coordinates in target coordinate reference system.
        """
        return rasterio.warp.transform_bounds(
            self.crs, dst_crs, *self.bounds(recalc=recalc), densify_pts=densify_pts
        )

    def write_gcps(
        self,
        gcps: Iterable[GroundControlPoint],
        gcp_crs: Any,
        *,
        grid_mapping_name: Optional[str] = None,
        inplace: bool = False,
    ) -> Union[xarray.Dataset, xarray.DataArray]:
        """
        Write the GroundControlPoints to the dataset.

        https://rasterio.readthedocs.io/en/latest/topics/georeferencing.html#ground-control-points

        Parameters
        ----------
        gcp: list of :obj:`rasterio.control.GroundControlPoint`
            The Ground Control Points to integrate to the dataset.
        gcp_crs: str, :obj:`rasterio.crs.CRS`, or dict
            Coordinate reference system for the GCPs.
        grid_mapping_name: str, optional
            Name of the grid_mapping coordinate to store the GCPs information in.
            Default is the grid_mapping name of the dataset.
        inplace: bool, optional
            If True, it will write to the existing dataset. Default is False.

        Returns
        -------
        :obj:`xarray.Dataset` | :obj:`xarray.DataArray`:
            Modified dataset with Ground Control Points written.
        """
        grid_mapping_name = (
            self.grid_mapping if grid_mapping_name is None else grid_mapping_name
        )
        data_obj = self._get_obj(inplace=True)

        if gcp_crs:
            data_obj = data_obj.rio.write_crs(
                gcp_crs, grid_mapping_name=grid_mapping_name, inplace=inplace
            )
        try:
            grid_map_attrs = data_obj.coords[grid_mapping_name].attrs.copy()
        except KeyError:
            data_obj.coords[grid_mapping_name] = xarray.Variable((), 0)
            grid_map_attrs = data_obj.coords[grid_mapping_name].attrs.copy()
        geojson_gcps = _convert_gcps_to_geojson(gcps)
        grid_map_attrs["gcps"] = json.dumps(geojson_gcps)
        data_obj.coords[grid_mapping_name].rio.set_attrs(grid_map_attrs, inplace=True)
        self._gcps = list(gcps)
        return data_obj

    def get_gcps(self) -> Optional[list[GroundControlPoint]]:
        """
        Get the GroundControlPoints from the dataset.

        https://rasterio.readthedocs.io/en/latest/topics/georeferencing.html#ground-control-points

        Returns
        -------
        list of :obj:`rasterio.control.GroundControlPoint` or None
            The Ground Control Points from the dataset or None if not applicable
        """
        if self._gcps is not None:
            return self._gcps
        try:
            geojson_gcps = json.loads(self._obj.coords[self.grid_mapping].attrs["gcps"])
        except (KeyError, AttributeError):
            return None

        def _parse_gcp(gcp) -> GroundControlPoint:
            x, y, *z = gcp["geometry"]["coordinates"]
            z = z[0] if z else None
            return GroundControlPoint(
                x=x,
                y=y,
                z=z,
                row=gcp["properties"]["row"],
                col=gcp["properties"]["col"],
                id=gcp["properties"]["id"],
                info=gcp["properties"]["info"],
            )

        self._gcps = [_parse_gcp(gcp) for gcp in geojson_gcps["features"]]
        return self._gcps

    def write_rpcs(
        self,
        rpcs: RPC,
        *,
        grid_mapping_name: Optional[str] = None,
        inplace: bool = False,
    ) -> xarray.Dataset | xarray.DataArray:
        """
        Write the Rational Polynomial Coefficients to the dataset.

        https://rasterio.readthedocs.io/en/latest/topics/georeferencing.html#rational-polynomial-coefficients

        Parameters
        ----------
        rpcs: :obj:`rasterio.rpc.RPC`
            The Rational Polynomial Coefficients to integrate to the dataset.
        grid_mapping_name: str, optional
            Name of the grid_mapping coordinate to store the RPCs information in.
            Default is the grid_mapping name of the dataset.
        inplace: bool, optional
            If True, it will write to the existing dataset. Default is False.

        Returns
        -------
        :obj:`xarray.Dataset` | :obj:`xarray.DataArray`:
            Modified dataset with Rational Polynomial Coefficients written.
        """
        grid_mapping_name = (
            self.grid_mapping if grid_mapping_name is None else grid_mapping_name
        )
        data_obj = self._get_obj(inplace=True)

        # RPC CRS is always 4326
        data_obj = data_obj.rio.write_crs(
            "epsg:4326", grid_mapping_name=grid_mapping_name, inplace=inplace
        )
        try:
            grid_map_attrs = data_obj.coords[grid_mapping_name].attrs.copy()
        except KeyError:
            data_obj.coords[grid_mapping_name] = xarray.Variable((), 0)
            grid_map_attrs = data_obj.coords[grid_mapping_name].attrs.copy()

        # Store the RPCCs
        grid_map_attrs["rpcs"] = json.dumps(rpcs.to_dict())
        data_obj.coords[grid_mapping_name].rio.set_attrs(grid_map_attrs, inplace=True)
        self._rpcs = rpcs

        return data_obj

    def get_rpcs(self) -> Optional[RPC]:
        """
        Get the Rational Polynomial Coefficients from the dataset.

        https://rasterio.readthedocs.io/en/latest/topics/georeferencing.html#rational-polynomial-coefficients

        Returns
        -------
        :obj:`rasterio.rpc.RPC` or None
            The Rational Polynomial Coefficients from the dataset or None if not applicable
        """
        if self._rpcs is not None:
            return self._rpcs
        try:
            json_rpcs = json.loads(self._obj.coords[self.grid_mapping].attrs["rpcs"])
        except (KeyError, AttributeError):
            return None

        self._rpcs = RPC(**json_rpcs)
        return self._rpcs



"""
This module allows you to merge xarray Datasets/DataArrays
geospatially with the `rasterio.merge` module.
"""

from collections.abc import Sequence
from typing import Callable, Optional, Union

import numpy
from rasterio.crs import CRS
from rasterio.merge import merge as _rio_merge
from xarray import DataArray, Dataset, IndexVariable

from rioxarray._spatial_utils import _get_nonspatial_coords, _make_coords


class RasterioDatasetDuck:
    """
    This class is to provide the attributes and methods necessary
    to make the :func:`rasterio.merge.merge` function think that
    the :obj:`xarray.DataArray` is a :obj:`rasterio.io.DatasetReader`.
    """

    # pylint: disable=too-many-instance-attributes

    def __init__(self, xds: DataArray):
        self._xds = xds
        self.crs = xds.rio.crs
        self.bounds = xds.rio.bounds(recalc=True)
        self.count = int(xds.rio.count)
        self.dtypes = [xds.dtype]
        self.name = xds.name
        self.nodatavals = [xds.rio.nodata]
        res = xds.rio.resolution(recalc=True)
        self.res = (abs(res[0]), abs(res[1]))
        self.transform = xds.rio.transform(recalc=True)
        # profile is only used for writing to a file.
        # This never happens with rioxarray merge.
        self.profile: dict = {}

    def colormap(self, *args, **kwargs) -> None:
        """
        colormap is only used for writing to a file.
        This never happens with rioxarray merge.
        """
        # pylint: disable=unused-argument
        return None

    def read(self, window, out_shape, *args, **kwargs) -> numpy.ma.MaskedArray:
        # pylint: disable=unused-argument
        """
        This method is meant to be used by the rasterio.merge.merge function.
        """
        data_window = self._xds.rio.isel_window(window)
        if data_window.shape != out_shape:
            # in this section, the data is geographically the same
            # however it is not the same dimensions as requested
            # so need to resample to the requested shape
            if len(out_shape) == 3:
                _, out_height, out_width = out_shape
            else:
                out_height, out_width = out_shape
            data_window = self._xds.rio.reproject(
                self._xds.rio.crs,
                transform=self.transform,
                shape=(out_height, out_width),
            )

        nodata = self.nodatavals[0]
        mask = False
        fill_value = None
        if nodata is not None and numpy.isnan(nodata):
            mask = numpy.isnan(data_window)
        elif nodata is not None:
            mask = data_window == nodata
            fill_value = nodata

        # make sure the returned shape matches
        # the expected shape. This can be the case
        # when the xarray dataset was squeezed to 2D beforehand
        if len(out_shape) == 3 and len(data_window.shape) == 2:
            data_window = data_window.values.reshape((1, out_height, out_width))

        return numpy.ma.array(
            data_window, mask=mask, fill_value=fill_value, dtype=self.dtypes[0]
        )


def merge_arrays(
    dataarrays: Sequence[DataArray],
    *,
    bounds: Optional[tuple] = None,
    res: Optional[tuple] = None,
    nodata: Optional[float] = None,
    precision: Optional[float] = None,
    method: Union[str, Callable, None] = None,
    crs: Optional[CRS] = None,
    parse_coordinates: bool = True,
) -> DataArray:
    """
    Merge data arrays geospatially.

    Uses :func:`rasterio.merge.merge`

    .. versionadded:: 0.2 crs

    Parameters
    ----------
    dataarrays: list[xarray.DataArray]
        List of multiple xarray.DataArray with all geo attributes.
        The first one is assumed to have the same
        CRS, dtype, and dimensions as the others in the array.
    bounds: tuple, optional
        Bounds of the output image (left, bottom, right, top).
        If not set, bounds are determined from bounds of input DataArrays.
    res: tuple, optional
        Output resolution in units of coordinate reference system.
        If not set, the resolution of the first DataArray is used.
        If a single value is passed, output pixels will be square.
    nodata: float, optional
        nodata value to use in output file.
        If not set, uses the nodata value in the first input DataArray.
    precision: float, optional
        Number of decimal points of precision when computing inverse transform.
    method: str or callable, optional
        See :func:`rasterio.merge.merge` for details.
    crs: rasterio.crs.CRS, optional
        Output CRS. If not set, the CRS of the first DataArray is used.
    parse_coordinates: bool, optional
        If False, it will disable loading spatial coordinates.

    Returns
    -------
    :obj:`xarray.DataArray`:
        The geospatially merged data.
    """
    input_kwargs = {
        "bounds": bounds,
        "res": res,
        "nodata": nodata,
        "precision": precision,
        "method": method,
    }

    if crs is None:
        crs = dataarrays[0].rio.crs
    if res is None:
        res = tuple(abs(res_val) for res_val in dataarrays[0].rio.resolution())

    # prepare the duck arrays
    rioduckarrays = []
    for dataarray in dataarrays:
        da_res = tuple(abs(res_val) for res_val in dataarray.rio.resolution())
        if da_res != res or dataarray.rio.crs != crs:
            rioduckarrays.append(
                RasterioDatasetDuck(
                    dataarray.rio.reproject(dst_crs=crs, resolution=res)
                )
            )
        else:
            rioduckarrays.append(RasterioDatasetDuck(dataarray))

    # use rasterio to merge
    merged_data, merged_transform = _rio_merge(
        rioduckarrays,
        **{key: val for key, val in input_kwargs.items() if val is not None},
    )
    # generate merged data array
    representative_array = rioduckarrays[0]._xds
    if parse_coordinates:
        coords = _make_coords(
            src_data_array=representative_array,
            dst_affine=merged_transform,
            dst_width=merged_data.shape[-1],
            dst_height=merged_data.shape[-2],
        )
        if (
            representative_array.rio.x_dim != "x"
            and "x" in coords
            and coords["x"].ndim == 1
        ):
            coords[representative_array.rio.x_dim] = IndexVariable(
                representative_array.rio.x_dim, coords.pop("x")
            )
        if (
            representative_array.rio.y_dim != "y"
            and "y" in coords
            and coords["y"].ndim == 1
        ):
            coords[representative_array.rio.y_dim] = IndexVariable(
                representative_array.rio.y_dim, coords.pop("y")
            )
    else:
        coords = _get_nonspatial_coords(representative_array)

    # make sure the output merged data shape is 2D if the
    # original data was 2D. this can happen if the
    # xarray datasarray was squeezed.
    if len(merged_data.shape) == 3 and len(representative_array.shape) == 2:
        merged_data = merged_data.squeeze()

    xda = DataArray(
        name=representative_array.name,
        data=merged_data,
        coords=coords,
        dims=tuple(representative_array.dims),
        attrs=representative_array.attrs,
    )
    xda.encoding = representative_array.encoding.copy()
    xda.rio.write_nodata(
        nodata if nodata is not None else representative_array.rio.nodata, inplace=True
    )
    xda.rio.write_crs(
        representative_array.rio.crs,
        grid_mapping_name=representative_array.rio.grid_mapping,
        inplace=True,
    )
    xda.rio.write_transform(
        merged_transform,
        grid_mapping_name=representative_array.rio.grid_mapping,
        inplace=True,
    )
    return xda


def merge_datasets(
    datasets: Sequence[Dataset],
    *,
    bounds: Optional[tuple] = None,
    res: Optional[tuple] = None,
    nodata: Optional[float] = None,
    precision: Optional[float] = None,
    method: Union[str, Callable, None] = None,
    crs: Optional[CRS] = None,
) -> Dataset:
    """
    Merge datasets geospatially.

    Uses :func:`rasterio.merge.merge`

    .. versionadded:: 0.2 crs

    Parameters
    ----------
    datasets: list[xarray.Dataset]
        List of multiple xarray.Dataset with all geo attributes.
        The first one is assumed to have the same
        CRS, dtype, dimensions, and data_vars as the others in the array.
    bounds: tuple, optional
        Bounds of the output image (left, bottom, right, top).
        If not set, bounds are determined from bounds of input Dataset.
    res: tuple, optional
        Output resolution in units of coordinate reference system.
        If not set, the resolution of the first Dataset is used.
        If a single value is passed, output pixels will be square.
    nodata: float, optional
        nodata value to use in output file.
        If not set, uses the nodata value in the first input Dataset.
    precision: float, optional
        Number of decimal points of precision when computing inverse transform.
    method: str or callable, optional
        See rasterio docs.
    crs: rasterio.crs.CRS, optional
        Output CRS. If not set, the CRS of the first DataArray is used.

    Returns
    -------
    :obj:`xarray.Dataset`:
        The geospatially merged data.
    """

    representative_ds = datasets[0]
    merged_data = {}
    for iii, data_var in enumerate(representative_ds.data_vars):
        merged_data[data_var] = merge_arrays(
            [dataset[data_var] for dataset in datasets],
            bounds=bounds,
            res=res,
            nodata=nodata,
            precision=precision,
            method=method,
            crs=crs,
            parse_coordinates=iii == 0,
        )
    data_var = list(representative_ds.data_vars)[0]
    xds = Dataset(
        merged_data,
        attrs=representative_ds.attrs,
    )
    xds.rio.write_crs(merged_data[data_var].rio.crs, inplace=True)
    return xds



"""This module defines the common interface for convention implementations
and provides helpers for selecting conventions.
"""

from typing import Optional, Protocol, Union

import rasterio.crs
import xarray
from affine import Affine


class ConventionProtocol(Protocol):
    """Protocol defining the interface for convention modules."""

    @classmethod
    def read_crs(
        cls, obj: Union[xarray.Dataset, xarray.DataArray]
    ) -> Optional[rasterio.crs.CRS]:
        """Read CRS from the object using this convention."""

    @classmethod
    def read_transform(
        cls, obj: Union[xarray.Dataset, xarray.DataArray]
    ) -> Optional[Affine]:
        """Read transform from the object using this convention."""

    @classmethod
    def read_spatial_dimensions(
        cls,
        obj: Union[xarray.Dataset, xarray.DataArray],
    ) -> Optional[tuple[str, str]]:
        """Read spatial dimensions (y_dim, x_dim) from the object using this convention."""

    @classmethod
    def write_crs(
        cls,
        obj: Union[xarray.Dataset, xarray.DataArray],
        *,
        crs: rasterio.crs.CRS,
        **kwargs,
    ) -> Union[xarray.Dataset, xarray.DataArray]:
        """Write CRS to the object using this convention."""

    @classmethod
    def write_transform(
        cls,
        obj: Union[xarray.Dataset, xarray.DataArray],
        *,
        transform: Affine,
        **kwargs,
    ) -> Union[xarray.Dataset, xarray.DataArray]:
        """Write transform to the object using this convention."""



from rioxarray._show_versions import (
    _get_deps_info,
    _get_main_info,
    _get_sys_info,
    show_versions,
)


def test_get_main_info():
    pyproj_info = _get_main_info()
    assert "rasterio" in pyproj_info
    assert "xarray" in pyproj_info
    assert "GDAL" in pyproj_info
    assert "PROJ" in pyproj_info
    assert "GEOS" in pyproj_info
    assert "PROJ DATA" in pyproj_info
    assert "GDAL DATA" in pyproj_info


def test_get_sys_info():
    sys_info = _get_sys_info()

    assert "python" in sys_info
    assert "executable" in sys_info
    assert "machine" in sys_info


def test_get_deps_info():
    deps_info = _get_deps_info()

    assert "scipy" in deps_info
    assert "pyproj" in deps_info


def test_show_versions_with_proj(capsys):
    show_versions()
    out, err = capsys.readouterr()
    assert "System" in out
    assert "python" in out
    assert "GDAL" in out
    assert "rioxarray" in out
    assert "Other python deps" in out



"""Integration tests for reading Zarr conventions."""
import numpy as np
import pyproj
import xarray as xr
from affine import Affine
from rasterio.crs import CRS

import rioxarray  # noqa: F401
from rioxarray import set_options
from rioxarray._convention import zarr
from rioxarray.enum import Convention


def _create_zarr_array_with_proj():
    """Create a DataArray with Zarr proj: convention attributes."""
    data = xr.DataArray(
        np.random.rand(10, 20),
        dims=["y", "x"],
        coords={
            "y": np.arange(10),
            "x": np.arange(20),
        },
    )
    data.attrs["zarr_conventions"] = [zarr.PROJ_CONVENTION]
    data.attrs["proj:wkt2"] = CRS.from_epsg(4326).to_wkt()
    return data


def _create_zarr_array_with_spatial():
    """Create a DataArray with Zarr spatial: convention attributes."""
    data = xr.DataArray(
        np.random.rand(10, 20),
        dims=["lat", "lon"],
        coords={
            "lat": np.arange(10),
            "lon": np.arange(20),
        },
    )
    data.attrs["zarr_conventions"] = [zarr.SPATIAL_CONVENTION]
    data.attrs["spatial:transform"] = [1.0, 0.0, 100.0, 0.0, -1.0, 200.0]
    data.attrs["spatial:dimensions"] = ["lat", "lon"]
    return data


def _create_zarr_array_with_both():
    """Create a DataArray with both Zarr conventions."""
    data = xr.DataArray(
        np.random.rand(10, 20),
        dims=["lat", "lon"],
        coords={
            "lat": np.arange(10),
            "lon": np.arange(20),
        },
    )
    data.attrs["zarr_conventions"] = [zarr.PROJ_CONVENTION, zarr.SPATIAL_CONVENTION]
    data.attrs["proj:wkt2"] = CRS.from_epsg(32618).to_wkt()
    data.attrs["spatial:transform"] = [10.0, 0.0, 500000.0, 0.0, -10.0, 4500000.0]
    data.attrs["spatial:dimensions"] = ["lat", "lon"]
    return data


def test_read_crs_from_zarr_convention():
    """Test reading CRS from DataArray with Zarr proj: convention."""
    data = _create_zarr_array_with_proj()

    crs = data.rio.crs
    assert crs is not None
    assert crs == CRS.from_epsg(4326)


def test_read_crs_from_zarr_convention__with_setting():
    """Test reading CRS with Convention.ZARR setting."""
    data = _create_zarr_array_with_proj()

    with set_options(convention=Convention.ZARR):
        crs = data.rio.crs
        assert crs is not None
        assert crs == CRS.from_epsg(4326)


def test_read_transform_from_zarr_convention():
    """Test reading transform from DataArray with Zarr spatial: convention."""
    data = _create_zarr_array_with_spatial()

    # Access transform via rio accessor
    # Check the cached version reads from Zarr spatial:transform
    cached = data.rio._cached_transform()
    assert cached is not None
    assert cached == Affine(1.0, 0.0, 100.0, 0.0, -1.0, 200.0)


def test_read_spatial_dimensions_from_zarr_convention():
    """Test reading spatial dimensions from Zarr spatial: convention."""
    data = _create_zarr_array_with_spatial()

    assert data.rio.x_dim == "lon"
    assert data.rio.y_dim == "lat"


def test_read_both_conventions():
    """Test reading from DataArray with both Zarr conventions."""
    data = _create_zarr_array_with_both()

    # CRS from proj:
    crs = data.rio.crs
    assert crs is not None
    assert crs == CRS.from_epsg(32618)

    # Transform from spatial:
    cached = data.rio._cached_transform()
    assert cached is not None
    assert cached == Affine(10.0, 0.0, 500000.0, 0.0, -10.0, 4500000.0)

    # Dimensions from spatial:
    assert data.rio.x_dim == "lon"
    assert data.rio.y_dim == "lat"


def test_fallback_zarr_to_cf():
    """Test that CF convention is tried as fallback when Zarr not found."""
    # Create data with CF convention
    data = xr.DataArray(
        np.random.rand(10, 20),
        dims=["y", "x"],
        coords={
            "y": np.arange(10),
            "x": np.arange(20),
        },
    )
    data.coords["spatial_ref"] = xr.Variable((), 0)
    data.coords["spatial_ref"].attrs["spatial_ref"] = "EPSG:4326"

    # Even with Zarr preference, should fall back to CF
    with set_options(convention=Convention.ZARR):
        crs = data.rio.crs
        assert crs is not None
        assert crs == CRS.from_epsg(4326)


def test_fallback_cf_to_zarr():
    """Test that Zarr convention is tried as fallback when CF not found."""
    # Create data with Zarr convention only
    data = _create_zarr_array_with_proj()

    # With CF preference (default), should fall back to Zarr
    crs = data.rio.crs
    assert crs is not None
    assert crs == CRS.from_epsg(4326)


def test_priority_zarr_over_cf():
    """Test that Zarr convention takes priority when setting is Zarr."""
    # Create data with both conventions (different CRS values)
    data = xr.DataArray(
        np.random.rand(10, 20),
        dims=["y", "x"],
        coords={
            "y": np.arange(10),
            "x": np.arange(20),
        },
    )
    # CF convention
    data.coords["spatial_ref"] = xr.Variable((), 0)
    data.coords["spatial_ref"].attrs["spatial_ref"] = "EPSG:4326"

    # Zarr convention (different CRS)
    data.attrs["zarr_conventions"] = [zarr.PROJ_CONVENTION]
    data.attrs["proj:wkt2"] = CRS.from_epsg(32618).to_wkt()

    # With Zarr setting, should prefer Zarr CRS
    with set_options(convention=Convention.ZARR):
        crs = data.rio.crs
        assert crs is not None
        assert crs == CRS.from_epsg(32618)

    # Reset to check default
    data2 = data.copy(deep=True)
    data2.rio._crs = None  # Reset cached CRS

    # With default setting (CF priority), should prefer CF CRS
    crs = data2.rio.crs
    assert crs is not None
    assert crs == CRS.from_epsg(4326)


def test_read_proj_code():
    """Test reading CRS from proj:code attribute."""
    data = xr.DataArray(
        np.random.rand(10, 20),
        dims=["y", "x"],
    )
    data.attrs["zarr_conventions"] = [zarr.PROJ_CONVENTION]
    data.attrs["proj:code"] = "EPSG:32618"

    crs = data.rio.crs
    assert crs is not None
    assert crs == CRS.from_epsg(32618)


def test_read_proj_projjson():
    """Test reading CRS from proj:projjson attribute."""
    data = xr.DataArray(
        np.random.rand(10, 20),
        dims=["y", "x"],
    )
    data.attrs["zarr_conventions"] = [zarr.PROJ_CONVENTION]
    data.attrs["proj:projjson"] = pyproj.CRS.from_epsg(4326).to_json_dict()

    crs = data.rio.crs
    assert crs is not None
    assert crs == CRS.from_epsg(4326)



"""
This module is an extension for xarray to provide rasterio capabilities
to xarray datasets.
"""

import os
from collections.abc import Iterable, Mapping
from typing import Any, Literal, Optional, Union
from uuid import uuid4

import numpy
import rasterio.crs
import xarray
from affine import Affine
from rasterio.enums import Resampling

from rioxarray._options import SKIP_MISSING_SPATIAL_DIMS, get_option
from rioxarray.exceptions import MissingSpatialDimensionError, RioXarrayError
from rioxarray.rioxarray import XRasterBase, _get_spatial_dims


@xarray.register_dataset_accessor("rio")
class RasterDataset(XRasterBase):
    """This is the GIS extension for :class:`xarray.Dataset`"""

    @property
    def vars(self) -> list:
        """list: Returns non-coordinate varibles"""
        return list(self._obj.data_vars)

    @property
    def crs(self) -> Optional[rasterio.crs.CRS]:
        """:obj:`rasterio.crs.CRS`:
        Retrieve projection from `xarray.Dataset`
        """
        if self._crs is not None:
            return None if self._crs is False else self._crs
        self._crs = super().crs
        if self._crs is not None:
            return self._crs
        # ensure all the CRS of the variables are the same
        crs_list = []
        for var in self.vars:
            if self._obj[var].rio.crs is not None:
                crs_list.append(self._obj[var].rio.crs)
        try:
            crs = crs_list[0]
        except IndexError:
            crs = None
        if crs is None:
            self._crs = False
            return None
        if all(crs_i == crs for crs_i in crs_list):
            self._crs = crs
        else:
            raise RioXarrayError(f"CRS in DataArrays differ in the Dataset: {crs_list}")
        return self._crs

    def reproject(
        self,
        dst_crs: Any,
        *,
        resolution: Optional[Union[float, tuple[float, float]]] = None,
        shape: Optional[tuple[int, int]] = None,
        transform: Optional[Affine] = None,
        resampling: Optional[Union[Resampling | str]] = Resampling.nearest,
        nodata: Optional[float] = None,
        **kwargs,
    ) -> xarray.Dataset:
        """
        Reproject :class:`xarray.Dataset` objects

        .. note:: Only 2D/3D arrays with dimensions 'x'/'y' are currently supported.
            Others are appended as is.
            Requires either a grid mapping variable with 'spatial_ref' or
            a 'crs' attribute to be set containing a valid CRS.
            If using a WKT (e.g. from spatiareference.org), make sure it is an OGC WKT.

        .. note:: To re-project with dask, see
            `odc-geo <https://odc-geo.readthedocs.io/>`__ &
            `pyresample <https://pyresample.readthedocs.io/>`__.

        .. versionadded:: 0.0.27 shape
        .. versionadded:: 0.0.28 transform
        .. versionadded:: 0.5.0 nodata, kwargs

        Parameters
        ----------
        dst_crs: str
            OGC WKT string or Proj.4 string.
        resolution: float or tuple(float, float), optional
            Size of a destination pixel in destination projection units
            (e.g. degrees or metres).
        shape: tuple(int, int), optional
            Shape of the destination in pixels (dst_height, dst_width). Cannot be used
            together with resolution.
        transform: Affine, optional
            The destination transform.
        resampling: rasterio.enums.Resampling or str, optional
            See :func:`rasterio.warp.reproject` for more details. Will accept a string
            representation of the enum (e.g. "nearest" > Resampling.nearest).
        nodata: float, optional
            The nodata value used to initialize the destination;
            it will remain in all areas not covered by the reprojected source.
            Defaults to the nodata value of the source image if none provided
            and exists or attempts to find an appropriate value by dtype.
        **kwargs: dict
            Additional keyword arguments to pass into :func:`rasterio.warp.reproject`.
            To override:
            - src_transform: `rio.write_transform`
            - src_crs: `rio.write_crs`
            - src_nodata: `rio.write_nodata`

        Returns
        --------
        :class:`xarray.Dataset`:
            The reprojected Dataset.
        """
        resampled_dataset = xarray.Dataset(attrs=self._obj.attrs)
        for var in self.vars:
            try:
                x_dim, y_dim = _get_spatial_dims(self._obj, var=var)
                resampled_dataset[var] = (
                    self._obj[var]
                    .rio.set_spatial_dims(x_dim=x_dim, y_dim=y_dim, inplace=True)
                    .rio.reproject(
                        dst_crs,
                        resolution=resolution,
                        shape=shape,
                        transform=transform,
                        resampling=resampling,
                        nodata=nodata,
                        **kwargs,
                    )
                )
            except MissingSpatialDimensionError:
                if len(self._obj[var].dims) >= 2 and not get_option(
                    SKIP_MISSING_SPATIAL_DIMS
                ):
                    raise
                resampled_dataset[var] = self._obj[var].copy()
        return resampled_dataset

    def reproject_match(
        self,
        match_data_array: Union[xarray.DataArray, xarray.Dataset],
        *,
        resampling: Optional[Union[Resampling | str]] = Resampling.nearest,
        **reproject_kwargs,
    ) -> xarray.Dataset:
        """
        Reproject a Dataset object to match the resolution, projection,
        and region of another DataArray.

        .. note:: Only 2D/3D arrays with dimensions 'x'/'y' are currently supported.
            Others are appended as is.
            Requires either a grid mapping variable with 'spatial_ref' or
            a 'crs' attribute to be set containing a valid CRS.
            If using a WKT (e.g. from spatiareference.org), make sure it is an OGC WKT.

        .. versionadded:: 0.9 reproject_kwargs

        Parameters
        ----------
        match_data_array: :obj:`xarray.DataArray` | :obj:`xarray.Dataset`
            Dataset with the target resolution and projection.
        resampling: rasterio.enums.Resampling or str, optional
            See :func:`rasterio.warp.reproject` for more details. Will accept a string
            representation of the enum (e.g. "nearest" > Resampling.nearest).
        **reproject_kwargs:
            Other options to pass to :meth:`rioxarray.raster_dataset.RasterDataset.reproject`

        Returns
        --------
        :obj:`xarray.Dataset`:
            Contains the data from the src_data_array,
            reprojected to match match_data_array.
        """
        resampled_dataset = xarray.Dataset(attrs=self._obj.attrs)
        for var in self.vars:
            try:
                x_dim, y_dim = _get_spatial_dims(self._obj, var=var)
                resampled_dataset[var] = (
                    self._obj[var]
                    .rio.set_spatial_dims(x_dim=x_dim, y_dim=y_dim, inplace=True)
                    .rio.reproject_match(
                        match_data_array, resampling=resampling, **reproject_kwargs
                    )
                )
            except MissingSpatialDimensionError:
                if len(self._obj[var].dims) >= 2 and not get_option(
                    SKIP_MISSING_SPATIAL_DIMS
                ):
                    raise
                resampled_dataset[var] = self._obj[var].copy()
        return resampled_dataset

    def pad_box(
        self,
        minx: float,
        miny: float,
        maxx: float,
        maxy: float,
        *,
        constant_values: Union[
            float, tuple[int, int], Mapping[Any, tuple[int, int]], None
        ] = None,
    ) -> xarray.Dataset:
        """Pad the :class:`xarray.Dataset` to a bounding box.

        .. warning:: Only works if all variables in the dataset have the
                     same coordinates.

        .. warning:: Pads variables that have dimensions 'x'/'y'. Others are appended as is.

        Parameters
        ----------
        minx: float
            Minimum bound for x coordinate.
        miny: float
            Minimum bound for y coordinate.
        maxx: float
            Maximum bound for x coordinate.
        maxy: float
            Maximum bound for y coordinate.
        constant_values: scalar, tuple or mapping of hashable to tuple
            The value used for padding. If None, nodata will be used if it is
            set, and numpy.nan otherwise.

        Returns
        -------
        :obj:`xarray.Dataset`:
            The padded object.
        """
        padded_dataset = xarray.Dataset(attrs=self._obj.attrs)
        for var in self.vars:
            try:
                x_dim, y_dim = _get_spatial_dims(self._obj, var=var)
                padded_dataset[var] = (
                    self._obj[var]
                    .rio.set_spatial_dims(x_dim=x_dim, y_dim=y_dim, inplace=True)
                    .rio.pad_box(
                        minx, miny, maxx, maxy, constant_values=constant_values
                    )
                )
            except MissingSpatialDimensionError:
                if len(self._obj[var].dims) >= 2 and not get_option(
                    SKIP_MISSING_SPATIAL_DIMS
                ):
                    raise
                padded_dataset[var] = self._obj[var].copy()
        return padded_dataset.rio.set_spatial_dims(
            x_dim=self.x_dim, y_dim=self.y_dim, inplace=True
        )

    def clip_box(
        self,
        minx: float,
        miny: float,
        maxx: float,
        maxy: float,
        *,
        auto_expand: Union[bool, int] = False,
        auto_expand_limit: int = 3,
        crs: Optional[Any] = None,
        allow_one_dimensional_raster: bool = False,
    ) -> xarray.Dataset:
        """Clip the :class:`xarray.Dataset` by a bounding box in dimensions 'x'/'y'.

        .. warning:: Clips variables that have dimensions 'x'/'y'. Others are appended as is.

        .. versionadded:: 0.12 crs
        .. versionadded:: 0.16 allow_one_dimensional_raster

        Parameters
        ----------
        minx: float
            Minimum bound for x coordinate.
        miny: float
            Minimum bound for y coordinate.
        maxx: float
            Maximum bound for x coordinate.
        maxy: float
            Maximum bound for y coordinate.
        auto_expand: bool
            If True, it will expand clip search if only 1D raster found with clip.
        auto_expand_limit: int
            maximum number of times the clip will be retried before raising
            an exception.
        crs: :obj:`rasterio.crs.CRS`, optional
            The CRS of the bounding box. Default is to assume it is the same
            as the dataset.
        allow_one_dimensional_raster: bool, optional
            If True, allow clipping to/from a one dimensional raster.

        Returns
        -------
        Dataset:
            The clipped object.
        """
        clipped_dataset = xarray.Dataset(attrs=self._obj.attrs)
        for var in self.vars:
            try:
                x_dim, y_dim = _get_spatial_dims(self._obj, var=var)
                clipped_dataset[var] = (
                    self._obj[var]
                    .rio.set_spatial_dims(x_dim=x_dim, y_dim=y_dim, inplace=True)
                    .rio.clip_box(
                        minx,
                        miny,
                        maxx,
                        maxy,
                        auto_expand=auto_expand,
                        auto_expand_limit=auto_expand_limit,
                        crs=crs,
                        allow_one_dimensional_raster=allow_one_dimensional_raster,
                    )
                )
            except MissingSpatialDimensionError:
                if len(self._obj[var].dims) >= 2 and not get_option(
                    SKIP_MISSING_SPATIAL_DIMS
                ):
                    raise
                clipped_dataset[var] = self._obj[var].copy()
        return clipped_dataset.rio.set_spatial_dims(
            x_dim=self.x_dim, y_dim=self.y_dim, inplace=True
        )

    def clip(
        self,
        geometries: Iterable,
        crs: Optional[Any] = None,
        *,
        all_touched: bool = False,
        drop: bool = True,
        invert: bool = False,
        from_disk: bool = False,
    ) -> xarray.Dataset:
        """
        Crops a :class:`xarray.Dataset` by geojson like geometry dicts in dimensions 'x'/'y'.

        .. warning:: Clips variables that have dimensions 'x'/'y'. Others are appended as is.

        Powered by `rasterio.features.geometry_mask`.

        Examples:

            >>> geometry = ''' {"type": "Polygon",
            ...                 "coordinates": [
            ...                 [[-94.07955380199459, 41.69085871273774],
            ...                 [-94.06082436942204, 41.69103313774798],
            ...                 [-94.06063203899649, 41.67932439500822],
            ...                 [-94.07935807746362, 41.679150041277325],
            ...                 [-94.07955380199459, 41.69085871273774]]]}'''
            >>> cropping_geometries = [geojson.loads(geometry)]
            >>> xds = xarray.open_rasterio('cool_raster.tif')
            >>> cropped = xds.rio.clip(geometries=cropping_geometries, crs=4326)


        .. versionadded:: 0.2 from_disk

        Parameters
        ----------
        geometries: list
            A list of geojson geometry dicts.
        crs: :obj:`rasterio.crs.CRS`, optional
            The CRS of the input geometries. Default is to assume it is the same
            as the dataset.
        all_touched : boolean, optional
            If True, all pixels touched by geometries will be burned in.  If
            false, only pixels whose center is within the polygon or that
            are selected by Bresenham's line algorithm will be burned in.
        drop: bool, optional
            If True, drop the data outside of the extent of the mask geometries
            Otherwise, it will return the same raster with the data masked.
            Default is True.
        invert: boolean, optional
            If False, pixels that do not overlap shapes will be set as nodata.
            Otherwise, pixels that overlap the shapes will be set as nodata.
            False by default.
        from_disk: boolean, optional
            If True, it will clip from disk using rasterio.mask.mask if possible.
            This is beneficial when the size of the data is larger than memory.
            Default is False.

        Returns
        -------
        :obj:`xarray.Dataset`:
            The clipped object.
        """
        clipped_dataset = xarray.Dataset(attrs=self._obj.attrs)
        for var in self.vars:
            try:
                x_dim, y_dim = _get_spatial_dims(self._obj, var=var)
                clipped_dataset[var] = (
                    self._obj[var]
                    .rio.set_spatial_dims(x_dim=x_dim, y_dim=y_dim, inplace=True)
                    .rio.clip(
                        geometries,
                        crs=crs,
                        all_touched=all_touched,
                        drop=drop,
                        invert=invert,
                        from_disk=from_disk,
                    )
                )
            except MissingSpatialDimensionError:
                if len(self._obj[var].dims) >= 2 and not get_option(
                    SKIP_MISSING_SPATIAL_DIMS
                ):
                    raise
                clipped_dataset[var] = self._obj[var].copy()
        return clipped_dataset.rio.set_spatial_dims(
            x_dim=self.x_dim, y_dim=self.y_dim, inplace=True
        )

    def interpolate_na(
        self, method: Literal["linear", "nearest", "cubic"] = "nearest"
    ) -> xarray.Dataset:
        """
        This method uses `scipy.interpolate.griddata` to interpolate missing data.

        .. warning:: scipy is an optional dependency.

        .. warning:: Interpolates variables that have dimensions 'x'/'y'. Others are appended as is.

        Parameters
        ----------
        method: {'linear', 'nearest', 'cubic'}, optional
            The method to use for interpolation in `scipy.interpolate.griddata`.

        Returns
        -------
        :obj:`xarray.DataArray`:
             The interpolated object.
        """
        interpolated_dataset = xarray.Dataset(attrs=self._obj.attrs)
        for var in self.vars:
            try:
                x_dim, y_dim = _get_spatial_dims(self._obj, var=var)
                interpolated_dataset[var] = (
                    self._obj[var]
                    .rio.set_spatial_dims(x_dim=x_dim, y_dim=y_dim, inplace=True)
                    .rio.interpolate_na(method=method)
                )
            except MissingSpatialDimensionError:
                if len(self._obj[var].dims) >= 2 and not get_option(
                    SKIP_MISSING_SPATIAL_DIMS
                ):
                    raise
                interpolated_dataset[var] = self._obj[var].copy()
        return interpolated_dataset.rio.set_spatial_dims(
            x_dim=self.x_dim, y_dim=self.y_dim, inplace=True
        )

    def to_raster(
        self,
        raster_path: Union[str, os.PathLike],
        *,
        driver: Optional[str] = None,
        dtype: Optional[Union[str, numpy.dtype]] = None,
        tags: Optional[dict[str, str]] = None,
        windowed: bool = False,
        recalc_transform: bool = True,
        lock: Optional[bool] = None,
        compute: bool = True,
        **profile_kwargs,
    ) -> None:
        """
        Export the Dataset to a raster file. Only works with 2D data.

        ..versionadded:: 0.2 lock

        Parameters
        ----------
        raster_path: str
            The path to output the raster to.
        driver: str, optional
            The name of the GDAL/rasterio driver to use to export the raster.
            Default is "GTiff" if rasterio < 1.2 otherwise it will autodetect.
        dtype: str, optional
            The data type to write the raster to. Default is the datasets dtype.
        tags: dict, optional
            A dictionary of tags to write to the raster.
        windowed: bool, optional
            If True, it will write using the windows of the output raster.
            This is useful for loading data in chunks when writing. Does not
            do anything when writing with dask.
            Default is False.
        recalc_transform: bool, optional
            If False, it will write the raster with the cached transform from
            the dataset rather than recalculating it.
            Default is True.
        lock: boolean or Lock, optional
            Lock to use to write data using dask.
            If not supplied, it will use a single process for writing.
        compute: bool, optional
            If True and data is a dask array, then compute and save
            the data immediately. If False, return a dask Delayed object.
            Call ".compute()" on the Delayed object to compute the result
            later. Call ``dask.compute(delayed1, delayed2)`` to save
            multiple delayed files at once. Default is True.
        **profile_kwargs
            Additional keyword arguments to pass into writing the raster. The
            nodata, transform, crs, count, width, and height attributes
            are ignored.

        Returns
        -------
        :obj:`dask.Delayed` | :obj:`dask.Array` | None:
            If the data array is a dask array and compute
            is True. Otherwise None is returned.

        """
        # pylint: disable=too-many-locals
        variable_dim = f"band_{uuid4()}"
        data_array = self._obj.to_array(dim=variable_dim)
        # ensure raster metadata preserved
        attr_scales = []
        attr_offsets = []
        attr_nodatavals = []
        encoded_scales = []
        encoded_offsets = []
        encoded_nodatavals = []
        band_tags = []
        long_name = []
        for data_var in data_array[variable_dim].values:
            try:
                encoded_scales.append(self._obj[data_var].encoding["scale_factor"])
            except KeyError:
                attr_scales.append(self._obj[data_var].attrs.get("scale_factor", 1.0))
            try:
                encoded_offsets.append(self._obj[data_var].encoding["add_offset"])
            except KeyError:
                attr_offsets.append(self._obj[data_var].attrs.get("add_offset", 0.0))
            long_name.append(self._obj[data_var].attrs.get("long_name", data_var))
            if self._obj[data_var].rio.encoded_nodata is not None:
                encoded_nodatavals.append(self._obj[data_var].rio.encoded_nodata)
            else:
                attr_nodatavals.append(self._obj[data_var].rio.nodata)
            band_tags.append(self._obj[data_var].attrs.copy())
        if encoded_scales:
            data_array.encoding["scales"] = encoded_scales
        else:
            data_array.attrs["scales"] = attr_scales
        if encoded_offsets:
            data_array.encoding["offsets"] = encoded_offsets
        else:
            data_array.attrs["offsets"] = attr_offsets
        data_array.attrs["band_tags"] = band_tags
        data_array.attrs["long_name"] = long_name

        use_encoded_nodatavals = bool(encoded_nodatavals)
        nodatavals = encoded_nodatavals if use_encoded_nodatavals else attr_nodatavals
        nodata = nodatavals[0]
        if (
            all(nodataval == nodata for nodataval in nodatavals)
            or numpy.isnan(nodatavals).all()
        ):
            data_array.rio.write_nodata(
                nodata, inplace=True, encoded=use_encoded_nodatavals
            )
        else:
            raise RioXarrayError(
                "All nodata values must be the same when exporting to raster. "
                f"Current values: {attr_nodatavals}"
            )
        if self.crs is not None:
            data_array.rio.write_crs(self.crs, inplace=True)
        # write it to a raster
        return data_array.rio.set_spatial_dims(
            x_dim=self.x_dim,
            y_dim=self.y_dim,
            inplace=True,
        ).rio.to_raster(
            raster_path=raster_path,
            driver=driver,
            dtype=dtype,
            tags=tags,
            windowed=windowed,
            recalc_transform=recalc_transform,
            lock=lock,
            compute=compute,
            **profile_kwargs,
        )



"""
Helper functions for determining spatial attributes.
"""
import copy
import math
import warnings
from collections.abc import Hashable
from typing import Any, Iterable, Optional, Union

import numpy
import rasterio.mask
import rasterio.warp
import rasterio.windows
import xarray
from affine import Affine
from rasterio.control import GroundControlPoint
from rasterio.features import geometry_mask

from rioxarray.exceptions import MissingSpatialDimensionError

FILL_VALUE_NAMES = ("_FillValue", "missing_value", "fill_value", "nodata")
UNWANTED_RIO_ATTRS = ("nodatavals", "is_tiled", "res")
DEFAULT_GRID_MAP = "spatial_ref"

# DTYPE TO NODATA MAP
# Based on: https://github.com/OSGeo/gdal/blob/v3.12.1/swig/python/gdal-utils/osgeo_utils/gdal_calc.py#L49-L66
# And: https://github.com/rasterio/rasterio/blob/1.5.0/rasterio/dtypes.py#L91-L103
_NODATA_DTYPE_MAP = {
    1: 255,  # GDT_Byte
    2: 65535,  # GDT_UInt16
    3: -32768,  # GDT_Int16
    4: 4294967295,  # GDT_UInt32
    5: -2147483648,  # GDT_Int32
    6: numpy.nan,  # GDT_Float32
    7: numpy.nan,  # GDT_Float64
    8: None,  # GDT_CInt16
    9: None,  # GDT_CInt32
    10: numpy.nan,  # GDT_CFloat32
    11: numpy.nan,  # GDT_CFloat64
    12: 18446744073709551615,  # GDT_UInt64
    13: -9223372036854775808,  # GDT_Int64
    14: -128,  # GDT_Int8
    15: numpy.nan,  # GDT_Float16
    16: numpy.nan,  # GDT_CFloat16
}


def _affine_has_rotation(affine: Affine) -> bool:
    """
    Determine if the affine has rotation.

    Parameters
    ----------
    affine: :obj:`affine.Affine`
        The affine of the grid.

    Returns
    -------
    bool
    """
    return affine.b == affine.d != 0


def _resolution(affine: Affine) -> tuple[float, float]:
    """
    Determine if the resolution of the affine.
    If it has rotation, the sign of the resolution is lost.

    Based on: https://github.com/mapbox/rasterio/blob/6185a4e4ad72b5669066d2d5004bf46d94a6d298/rasterio/_base.pyx#L943-L951

    Parameters
    ----------
    affine: :obj:`affine.Affine`
        The affine of the grid.


    Returns
    --------
    x_resolution: float
        The X resolution of the affine.
    y_resolution: float
        The Y resolution of the affine.
    """
    if not _affine_has_rotation(affine):
        return affine.a, affine.e
    return (
        math.sqrt(affine.a**2 + affine.d**2),
        math.sqrt(affine.b**2 + affine.e**2),
    )


def affine_to_coords(
    affine: Affine, width: int, height: int, *, x_dim: str = "x", y_dim: str = "y"
) -> dict[str, numpy.ndarray]:
    """Generate 1d pixel centered coordinates from affine.

    Based on code from the xarray rasterio backend.

    Parameters
    ----------
    affine: :obj:`affine.Affine`
        The affine of the grid.
    width: int
        The width of the grid.
    height: int
        The height of the grid.
    x_dim: str, optional
        The name of the X dimension. Default is 'x'.
    y_dim: str, optional
        The name of the Y dimension. Default is 'y'.

    Returns
    -------
    dict: x and y coordinate arrays.

    """
    transform = affine * affine.translation(0.5, 0.5)
    if affine.is_rectilinear and not _affine_has_rotation(affine):
        x_coords, _ = transform * (numpy.arange(width), numpy.zeros(width))
        _, y_coords = transform * (numpy.zeros(height), numpy.arange(height))
    else:
        x_coords, y_coords = transform * numpy.meshgrid(
            numpy.arange(width),
            numpy.arange(height),
        )
    return {y_dim: y_coords, x_dim: x_coords}


def _generate_spatial_coords(
    *, affine: Affine, width: int, height: int
) -> dict[Hashable, Any]:
    """get spatial coords in new transform"""
    new_spatial_coords = affine_to_coords(affine, width, height)
    if new_spatial_coords["x"].ndim == 1:
        return {
            "x": xarray.IndexVariable("x", new_spatial_coords["x"]),
            "y": xarray.IndexVariable("y", new_spatial_coords["y"]),
        }
    return {
        "xc": (("y", "x"), new_spatial_coords["x"]),
        "yc": (("y", "x"), new_spatial_coords["y"]),
    }


def _get_nonspatial_coords(
    src_data_array: Union[xarray.DataArray, xarray.Dataset],
) -> dict[Hashable, Union[xarray.Variable, xarray.IndexVariable]]:
    coords: dict[Hashable, Union[xarray.Variable, xarray.IndexVariable]] = {}
    for coord in set(src_data_array.coords) - {
        src_data_array.rio.x_dim,
        src_data_array.rio.y_dim,
        DEFAULT_GRID_MAP,
    }:
        # skip 2D spatial coords
        if (
            src_data_array.rio.x_dim in src_data_array[coord].dims
            and src_data_array.rio.y_dim in src_data_array[coord].dims
        ):
            continue
        if src_data_array[coord].ndim == 1:
            coords[coord] = xarray.IndexVariable(
                src_data_array[coord].dims,
                src_data_array[coord].values,
                src_data_array[coord].attrs,
            )
        else:
            coords[coord] = xarray.Variable(
                src_data_array[coord].dims,
                src_data_array[coord].values,
                src_data_array[coord].attrs,
            )
    return coords


def _make_coords(
    *,
    src_data_array: Union[xarray.DataArray, xarray.Dataset],
    dst_affine: Affine,
    dst_width: int,
    dst_height: int,
    force_generate: bool = False,
) -> dict[Hashable, Any]:
    """Generate the coordinates of the new projected `xarray.DataArray`"""
    coords = _get_nonspatial_coords(src_data_array)
    if (
        force_generate
        or (
            src_data_array.rio.x_dim in src_data_array.coords
            and src_data_array.rio.y_dim in src_data_array.coords
        )
        or ("xc" in src_data_array.coords and "yc" in src_data_array.coords)
    ):
        new_coords = _generate_spatial_coords(
            affine=dst_affine, width=dst_width, height=dst_height
        )
        new_coords.update(coords)
        return new_coords
    return coords


def _get_data_var_message(obj: Union[xarray.DataArray, xarray.Dataset]) -> str:
    """
    Get message for named data variables.
    """
    try:
        return f" Data variable: {obj.name}" if obj.name else ""
    except AttributeError:
        return ""


def _get_spatial_dims(
    obj: Union[xarray.Dataset, xarray.DataArray], *, var: Union[Any, Hashable]
) -> tuple[str, str]:
    """
    Retrieve the spatial dimensions of the dataset
    """
    try:
        return obj[var].rio.x_dim, obj[var].rio.y_dim
    except MissingSpatialDimensionError as err:
        try:
            obj[var].rio.set_spatial_dims(
                x_dim=obj.rio.x_dim, y_dim=obj.rio.y_dim, inplace=True
            )
            return obj.rio.x_dim, obj.rio.y_dim
        except MissingSpatialDimensionError:
            raise err from None


def _has_spatial_dims(
    obj: Union[xarray.Dataset, xarray.DataArray], *, var: Union[Any, Hashable]
) -> bool:
    """
    Check to see if the variable in the Dataset has spatial dimensions
    """
    try:
        # pylint: disable=pointless-statement
        _get_spatial_dims(obj, var=var)
    except MissingSpatialDimensionError:
        return False
    return True


def _order_bounds(
    *,
    minx: float,
    miny: float,
    maxx: float,
    maxy: float,
    resolution_x: float,
    resolution_y: float,
) -> tuple[float, float, float, float]:
    """
    Make sure that the bounds are in the correct order
    """
    if resolution_y < 0:
        top = maxy
        bottom = miny
    else:
        top = miny
        bottom = maxy
    if resolution_x < 0:
        left = maxx
        right = minx
    else:
        left = minx
        right = maxx

    return left, bottom, right, top


def _convert_gcps_to_geojson(
    gcps: Iterable[GroundControlPoint],
) -> dict:
    """
    Convert GCPs to geojson.

    Parameters
    ----------
    gcps: The list of GroundControlPoint instances.

    Returns
    -------
    A FeatureCollection dict.
    """

    def _gcp_coordinates(gcp):
        if gcp.z is None:
            return [gcp.x, gcp.y]
        return [gcp.x, gcp.y, gcp.z]

    features = [
        {
            "type": "Feature",
            "properties": {
                "id": gcp.id,
                "info": gcp.info,
                "row": gcp.row,
                "col": gcp.col,
            },
            "geometry": {"type": "Point", "coordinates": _gcp_coordinates(gcp)},
        }
        for gcp in gcps
    ]
    return {"type": "FeatureCollection", "features": features}


def _convert_str_to_resampling(name: str) -> rasterio.warp.Resampling:
    """
    Convert from string to rasterio.warp.Resampling enum, raises ValueError on bad input.

    Parameters
    ----------
    name: str
        The string to convert.

    Returns
    -------
    :obj:`rasterio.warp.Resampling`
    """
    try:
        return getattr(rasterio.warp.Resampling, name.lower())
    except AttributeError:
        raise ValueError(f"Bad resampling parameter: {name}") from None


def _generate_attrs(
    *, src_data_array: xarray.DataArray, dst_nodata: Optional[float]
) -> dict[str, Any]:
    # add original attributes
    new_attrs = copy.deepcopy(src_data_array.attrs)
    # remove all nodata information
    for unwanted_attr in FILL_VALUE_NAMES + UNWANTED_RIO_ATTRS:
        new_attrs.pop(unwanted_attr, None)

    # add nodata information
    fill_value = (
        src_data_array.rio.nodata
        if src_data_array.rio.nodata is not None
        else dst_nodata
    )
    if src_data_array.rio.encoded_nodata is None and fill_value is not None:
        new_attrs["_FillValue"] = fill_value

    return new_attrs


def _add_attrs_proj(
    *, new_data_array: xarray.DataArray, src_data_array: xarray.DataArray
) -> xarray.DataArray:
    """Make sure attributes and projection correct"""
    # make sure dimension information is preserved
    if new_data_array.rio._x_dim is None:
        new_data_array.rio._x_dim = src_data_array.rio.x_dim
    if new_data_array.rio._y_dim is None:
        new_data_array.rio._y_dim = src_data_array.rio.y_dim

    # make sure attributes preserved
    new_attrs = _generate_attrs(src_data_array=src_data_array, dst_nodata=None)
    # remove fill value if it already exists in the encoding
    # this is for data arrays pulling the encoding from a
    # source data array instead of being generated anew.
    if "_FillValue" in new_data_array.encoding:
        new_attrs.pop("_FillValue", None)

    new_data_array.rio.set_attrs(new_attrs, inplace=True)

    # make sure projection added
    new_data_array.rio.write_crs(
        src_data_array.rio.crs,
        grid_mapping_name=src_data_array.rio.grid_mapping,
        inplace=True,
    )
    new_data_array.rio.write_coordinate_system(inplace=True)
    new_data_array.rio.write_transform(inplace=True)
    # make sure encoding added
    new_data_array.encoding = src_data_array.encoding.copy()
    return new_data_array


def _make_dst_affine(
    *,
    src_data_array: xarray.DataArray,
    src_crs: rasterio.crs.CRS,
    dst_crs: rasterio.crs.CRS,
    dst_resolution: Optional[Union[float, tuple[float, float]]] = None,
    dst_shape: Optional[tuple[float, float]] = None,
    **kwargs,
):
    """Determine the affine of the new projected `xarray.DataArray`"""
    src_bounds = ()
    if (
        "gcps" not in kwargs
        and "rpcs" not in kwargs
        and "src_geoloc_array" not in kwargs
    ):
        src_bounds = src_data_array.rio.bounds()
    src_height, src_width = src_data_array.rio.shape
    dst_height, dst_width = dst_shape if dst_shape is not None else (None, None)
    # pylint: disable=isinstance-second-argument-not-valid-type
    if isinstance(dst_resolution, Iterable):
        dst_resolution = tuple(abs(res_val) for res_val in dst_resolution)  # type: ignore
    elif dst_resolution is not None:
        dst_resolution = abs(dst_resolution)  # type: ignore

    for key, value in (
        ("resolution", dst_resolution),
        ("dst_height", dst_height),
        ("dst_width", dst_width),
    ):
        if value is not None:
            kwargs[key] = value
    dst_affine, dst_width, dst_height = rasterio.warp.calculate_default_transform(
        src_crs,
        dst_crs,
        src_width,
        src_height,
        *src_bounds,
        **kwargs,
    )
    return dst_affine, dst_width, dst_height


def _clip_from_disk(
    xds: xarray.DataArray,
    *,
    geometries: Iterable,
    all_touched: bool,
    drop: bool,
    invert: bool,
) -> Optional[xarray.DataArray]:
    """
    clip from disk if the file object is available
    """
    try:
        rio_dataset = xds.rio._manager.acquire()
    except AttributeError:
        warnings.warn("File object not available, clipping in-memory.")
        return None

    out_image, out_transform = rasterio.mask.mask(
        rio_dataset,
        geometries,
        all_touched=all_touched,
        invert=invert,
        crop=drop,
    )
    if xds.rio.encoded_nodata is not None and not numpy.isnan(xds.rio.encoded_nodata):
        out_image = out_image.astype(numpy.float64)
        out_image[out_image == xds.rio.encoded_nodata] = numpy.nan

    height, width = out_image.shape[-2:]
    cropped_ds = xarray.DataArray(
        name=xds.name,
        data=out_image,
        coords=_make_coords(
            src_data_array=xds,
            dst_affine=out_transform,
            dst_width=width,
            dst_height=height,
        ),
        dims=xds.dims,
        attrs=xds.attrs,
    )
    cropped_ds.encoding = xds.encoding
    return cropped_ds


def _clip_xarray(
    xds: xarray.DataArray,
    *,
    geometries: Iterable,
    all_touched: bool,
    drop: bool,
    invert: bool,
) -> xarray.DataArray:
    """
    clip the xarray DataArray
    """
    clip_mask_arr = geometry_mask(
        geometries=geometries,
        out_shape=(int(xds.rio.height), int(xds.rio.width)),
        transform=xds.rio.transform(recalc=True),
        invert=not invert,
        all_touched=all_touched,
    )
    clip_mask_xray = xarray.DataArray(
        clip_mask_arr,
        dims=(xds.rio.y_dim, xds.rio.x_dim),
    )
    cropped_ds = xds.where(clip_mask_xray)
    if drop:
        cropped_ds.rio.set_spatial_dims(
            x_dim=xds.rio.x_dim, y_dim=xds.rio.y_dim, inplace=True
        )
        cropped_ds = cropped_ds.rio.isel_window(
            rasterio.windows.get_data_window(
                numpy.ma.masked_array(clip_mask_arr, ~clip_mask_arr)
            )
        )
    if xds.rio.nodata is not None and not numpy.isnan(xds.rio.nodata):
        cropped_ds = cropped_ds.fillna(xds.rio.nodata)

    return cropped_ds.astype(xds.dtype)



import os.path

import pytest

from rioxarray.exceptions import RioXarrayError
from test.conftest import TEST_INPUT_DATA_DIR

xarray = pytest.importorskip("xarray", minversion="0.18")


@pytest.mark.parametrize(
    "kwargs", [{}, {"engine": "rasterio"}, {"decode_coords": "all"}]
)
def test_xarray_open_dataset(kwargs):
    cog_file = os.path.join(TEST_INPUT_DATA_DIR, "cog.tif")
    ds = xarray.open_dataset(cog_file, **kwargs)

    assert isinstance(ds, xarray.Dataset)
    assert "band_data" in ds.data_vars
    assert ds.data_vars["band_data"].shape == (1, 500, 500)
    assert "spatial_ref" not in ds.data_vars
    assert "spatial_ref" in ds.coords
    assert "grid_mapping" not in ds.data_vars["band_data"].attrs
    assert "grid_mapping" in ds.data_vars["band_data"].encoding
    assert "preferred_chunks" in ds.data_vars["band_data"].encoding


def test_xarray_open_dataset__drop_variables():
    input_file = os.path.join(
        TEST_INPUT_DATA_DIR, "MOD09GA.A2008296.h14v17.006.2015181011753.hdf"
    )

    with xarray.open_dataset(
        input_file,
        engine="rasterio",
        group="MODIS_Grid_500m_2D",
        drop_variables=[
            "sur_refl_b01_1",
            "sur_refl_b02_1",
            "sur_refl_b03_1",
            "sur_refl_b04_1",
            "sur_refl_b05_1",
            "sur_refl_b06_1",
            "sur_refl_b07_1",
        ],
    ) as rds:
        assert sorted(rds.data_vars) == [
            "QC_500m_1",
            "iobs_res_1",
            "num_observations_500m",
            "obscov_500m_1",
        ]


def test_open_multiple_resolution():
    with pytest.raises(
        RioXarrayError,
        match="Multiple resolution sets found. Use 'variable' or 'group' to filter.",
    ):
        xarray.open_dataset(
            os.path.join(
                TEST_INPUT_DATA_DIR, "MOD09GA.A2008296.h14v17.006.2015181011753.hdf"
            ),
            engine="rasterio",
            drop_variables="QC_500m_1",
        )


def test_xarray_open_dataset__invalid_decode_coords():
    cog_file = os.path.join(TEST_INPUT_DATA_DIR, "cog.tif")
    with pytest.raises(RioXarrayError):
        xarray.open_dataset(cog_file, decode_coords=False)



import numpy
import xarray

import rioxarray  # noqa: F401


def test_reproject_match__exact():
    """
    Based on: https://github.com/corteva/rioxarray/issues/298#issue-858511505
    """
    da1 = xarray.DataArray(
        numpy.random.rand(2, 3),
        [
            ("y", [-30.25, -30.3], {"id": 1, "random": "random"}),
            ("x", [149.8, 149.85, 149.9], {"id": 1, "random": "random"}),
        ],
    )
    da1.rio.write_crs(4326, inplace=True)
    da2 = xarray.DataArray(
        numpy.random.rand(4, 11),
        [
            (
                "longitude",
                [-30.25733604, -30.26550543, -30.27367483, -30.28184423],
                {"id": 2},
                {"rasterio_dtype": "int"},
            ),
            (
                "latitude",
                [
                    149.82193392,
                    149.83010332,
                    149.83827272,
                    149.84644211,
                    149.85461151,
                    149.86278091,
                    149.87095031,
                    149.87911971,
                    149.8872891,
                    149.8954585,
                    149.9036279,
                ],
                {"id": 2},
                {"rasterio_dtype": "int"},
            ),
        ],
    )
    da2.rio.write_crs(4326, inplace=True)
    resampled = da1.rio.reproject_match(da2)
    assert resampled.x.attrs == {
        "axis": "X",
        "long_name": "longitude",
        "standard_name": "longitude",
        "units": "degrees_east",
    }
    assert resampled.y.attrs == {
        "axis": "Y",
        "long_name": "latitude",
        "standard_name": "latitude",
        "units": "degrees_north",
    }
    numpy.testing.assert_array_equal(resampled.x, da2.longitude)
    numpy.testing.assert_array_equal(resampled.y, da2.latitude)



"""Enums for rioxarray."""
from enum import Enum


class Convention(Enum):
    """
    Supported geospatial metadata conventions.

    rioxarray supports conventions for storing geospatial metadata.
    Currently supported:

    - CF: Climate and Forecasts convention using grid_mapping coordinates
    - Zarr: Zarr spatial and proj conventions for cloud-native formats

    The convention can be set globally using set_options() or per-method
    using the convention parameter.

    Examples
    --------
    Set global convention:

    >>> import rioxarray
    >>> from rioxarray.enum import Convention
    >>> rioxarray.set_options(convention=Convention.CF)

    Use specific convention for a method:

    >>> from rioxarray.enum import Convention
    >>> data.rio.write_crs("EPSG:4326", convention=Convention.CF)

    See Also
    --------
    rioxarray.set_options : Set global options including convention

    References
    ----------
    .. [1] CF Conventions: https://github.com/cf-convention/cf-conventions
    .. [2] Zarr spatial convention: https://github.com/zarr-conventions/spatial
    .. [3] Zarr geo-proj convention: https://github.com/zarr-conventions/geo-proj
    """

    #: Climate and Forecasts convention (default)
    #: https://github.com/cf-convention/cf-conventions
    CF = "CF"

    #: Zarr spatial and proj conventions
    #: https://github.com/zarr-conventions/spatial
    #: https://github.com/zarr-conventions/geo-proj
    ZARR = "ZARR"



"""
Core convention methods for rioxarray.
"""

from typing import Optional, Union

import rasterio.crs
import xarray
from affine import Affine

from rioxarray._convention._base import ConventionProtocol
from rioxarray._convention.cf import CFConvention
from rioxarray._convention.zarr import ZarrConvention
from rioxarray._options import CONVENTION, get_option
from rioxarray.crs import crs_from_user_input
from rioxarray.enum import Convention

# Convention classes mapped by Convention enum
_CONVENTION_MODULES: dict[Convention, ConventionProtocol] = {
    Convention.CF: CFConvention,  # type: ignore[dict-item]
    Convention.ZARR: ZarrConvention,  # type: ignore[dict-item]
}


def _get_convention(convention: Convention | None) -> ConventionProtocol:
    """
    Get the convention module for writing.

    Parameters
    ----------
    convention : Convention enum value or None
        The convention to use. If None, uses the global default.

    Returns
    -------
    ConventionProtocol
        The module implementing the convention
    """
    if convention is None:
        convention = get_option(CONVENTION) or Convention.CF
    convention = Convention(convention)
    return _CONVENTION_MODULES[convention]


def read_crs_auto(
    obj: Union[xarray.Dataset, xarray.DataArray],
    **kwargs,
) -> Optional[rasterio.crs.CRS]:
    """
    Auto-detect and read CRS by trying convention readers.

    If a convention is set globally via set_options(), that convention
    is tried first for better performance. Then other conventions are
    tried as fallback.

    Parameters
    ----------
    obj : xarray.Dataset or xarray.DataArray
        Object to read CRS from
    **kwargs
        Convention-specific parameters (e.g., grid_mapping for CF)

    Returns
    -------
    rasterio.crs.CRS or None
        CRS object, or None if not found in any convention
    """
    # Try the configured convention first (if set)
    configured_convention = get_option(CONVENTION)
    if configured_convention is not None:
        result = _CONVENTION_MODULES[configured_convention].read_crs(obj, **kwargs)
        if result is not None:
            return result

    # Try all other conventions
    for conv_enum, convention in _CONVENTION_MODULES.items():
        if conv_enum == configured_convention:
            continue  # Already tried this one
        result = convention.read_crs(obj, **kwargs)
        if result is not None:
            return result

    # Legacy fallback: look in attrs for 'crs' (not part of any convention)
    try:
        return crs_from_user_input(obj.attrs["crs"])
    except KeyError:
        pass

    return None


def read_transform_auto(
    obj: Union[xarray.Dataset, xarray.DataArray],
    **kwargs,
) -> Optional[Affine]:
    """
    Auto-detect and read transform by trying convention readers.

    If a convention is set globally via set_options(), that convention
    is tried first for better performance. Then other conventions are
    tried as fallback.

    Parameters
    ----------
    obj : xarray.Dataset or xarray.DataArray
        Object to read transform from
    **kwargs
        Convention-specific parameters (e.g., grid_mapping for CF)

    Returns
    -------
    affine.Affine or None
        Transform object, or None if not found in any convention
    """
    # Try the configured convention first (if set)
    configured_convention = get_option(CONVENTION)
    if configured_convention is not None:
        result = _CONVENTION_MODULES[configured_convention].read_transform(
            obj, **kwargs
        )
        if result is not None:
            return result

    # Try all other conventions
    for conv_enum, convention in _CONVENTION_MODULES.items():
        if conv_enum == configured_convention:
            continue  # Already tried this one
        result = convention.read_transform(obj, **kwargs)
        if result is not None:
            return result

    # Legacy fallback: look in attrs for 'transform' (not part of any convention)
    try:
        return Affine(*obj.attrs["transform"][:6])
    except KeyError:
        pass

    return None


def read_spatial_dimensions_auto(
    obj: Union[xarray.Dataset, xarray.DataArray],
) -> Optional[tuple[str, str]]:
    """
    Auto-detect and read spatial dimensions by trying convention readers.

    If a convention is set globally via set_options(), that convention
    is tried first for better performance. Then other conventions are
    tried as fallback.

    Parameters
    ----------
    obj : xarray.Dataset or xarray.DataArray
        Object to read spatial dimensions from

    Returns
    -------
    tuple of (y_dim, x_dim) or None
        Tuple of dimension names, or None if not found in any convention
    """
    # Try the configured convention first (if set)
    configured_convention = get_option(CONVENTION)
    if configured_convention is not None:
        result = _CONVENTION_MODULES[configured_convention].read_spatial_dimensions(obj)
        if result is not None:
            return result

    # Try all other conventions
    for conv_enum, convention in _CONVENTION_MODULES.items():
        if conv_enum == configured_convention:
            continue  # Already tried this one
        result = convention.read_spatial_dimensions(obj)
        if result is not None:
            return result

    return None



"""Unit tests for the CF convention module."""
import numpy
import xarray
from affine import Affine
from rasterio.crs import CRS

from rioxarray._convention._core import read_crs_auto, read_transform_auto
from rioxarray._convention.cf import CFConvention


def test_read_crs__from_grid_mapping_spatial_ref():
    """Test reading CRS from grid_mapping coordinate's spatial_ref attribute."""
    data = xarray.DataArray(numpy.random.rand(10, 10), dims=["y", "x"])
    data.coords["spatial_ref"] = xarray.Variable((), 0)
    data.coords["spatial_ref"].attrs["spatial_ref"] = "EPSG:4326"

    crs = CFConvention.read_crs(data)
    assert crs is not None
    assert crs == CRS.from_epsg(4326)


def test_read_crs__from_grid_mapping_crs_wkt():
    """Test reading CRS from grid_mapping coordinate's crs_wkt attribute."""
    data = xarray.DataArray(numpy.random.rand(10, 10), dims=["y", "x"])
    data.coords["spatial_ref"] = xarray.Variable((), 0)
    data.coords["spatial_ref"].attrs["crs_wkt"] = CRS.from_epsg(4326).to_wkt()

    crs = CFConvention.read_crs(data)
    assert crs is not None
    assert crs == CRS.from_epsg(4326)


def test_read_crs__from_legacy_attrs():
    """Test reading CRS from object's attrs (legacy, not CF convention).

    The 'crs' attribute is not part of CF convention but is supported
    for backwards compatibility via the auto-detect method.
    """
    data = xarray.DataArray(numpy.random.rand(10, 10), dims=["y", "x"])
    data.attrs["crs"] = "EPSG:4326"

    # CF convention should NOT find this
    crs = CFConvention.read_crs(data)
    assert crs is None

    # Auto-detect should find it
    crs = read_crs_auto(data)
    assert crs is not None
    assert crs == CRS.from_epsg(4326)


def test_read_crs__from_legacy_attrs_with_missing_grid_mapping():
    """Test reading CRS from attrs when grid_mapping doesn't exist.

    This tests a common case where rioxarray's grid_mapping property returns
    "spatial_ref" as a default, but the coordinate doesn't actually exist.
    The CRS should still be found via auto-detect.
    """
    data = xarray.DataArray(numpy.random.rand(10, 10), dims=["y", "x"])
    data.attrs["crs"] = "EPSG:4326"

    # CF convention should NOT find this
    crs = CFConvention.read_crs(data)
    assert crs is None

    # Auto-detect should find it
    crs = read_crs_auto(data)
    assert crs is not None
    assert crs == CRS.from_epsg(4326)


def test_read_crs__not_found():
    """Test that None is returned when no CRS is found."""
    data = xarray.DataArray(numpy.random.rand(10, 10), dims=["y", "x"])

    crs = CFConvention.read_crs(data)
    assert crs is None


def test_read_transform__from_geotransform():
    """Test reading transform from GeoTransform attribute."""
    data = xarray.DataArray(numpy.random.rand(10, 10), dims=["y", "x"])
    data.coords["spatial_ref"] = xarray.Variable((), 0)
    # GeoTransform format: [c, a, b, f, d, e] (GDAL format)
    data.coords["spatial_ref"].attrs["GeoTransform"] = "0.0 1.0 0.0 10.0 0.0 -1.0"

    transform = CFConvention.read_transform(data)
    assert transform is not None
    assert transform == Affine(1.0, 0.0, 0.0, 0.0, -1.0, 10.0)


def test_read_transform__from_legacy_attrs():
    """Test reading transform from object's attrs (legacy, not CF convention).

    The 'transform' attribute is not part of CF convention but is supported
    for backwards compatibility via the auto-detect method.
    """
    data = xarray.DataArray(numpy.random.rand(10, 10), dims=["y", "x"])
    # Transform stored as list in attrs
    data.attrs["transform"] = [1.0, 0.0, 0.0, 0.0, -1.0, 10.0]

    # CF convention should NOT find this
    transform = CFConvention.read_transform(data)
    assert transform is None

    # Auto-detect should find it
    transform = read_transform_auto(data)
    assert transform is not None
    assert transform == Affine(1.0, 0.0, 0.0, 0.0, -1.0, 10.0)


def test_read_transform__not_found():
    """Test that None is returned when no transform is found."""
    data = xarray.DataArray(numpy.random.rand(10, 10), dims=["y", "x"])

    transform = CFConvention.read_transform(data)
    assert transform is None


def test_read_spatial_dimensions__xy():
    """Test detecting x/y dimension names."""
    data = xarray.DataArray(numpy.random.rand(10, 10), dims=["y", "x"])

    dims = CFConvention.read_spatial_dimensions(data)
    assert dims == ("y", "x")


def test_read_spatial_dimensions__lonlat():
    """Test detecting longitude/latitude dimension names."""
    data = xarray.DataArray(numpy.random.rand(10, 10), dims=["latitude", "longitude"])

    dims = CFConvention.read_spatial_dimensions(data)
    assert dims == ("latitude", "longitude")


def test_read_spatial_dimensions__cf_axis():
    """Test detecting dimensions from CF axis attributes."""
    data = xarray.DataArray(
        numpy.random.rand(10, 10),
        dims=["row", "col"],
        coords={
            "row": ("row", numpy.arange(10)),
            "col": ("col", numpy.arange(10)),
        },
    )
    data.coords["row"].attrs["axis"] = "Y"
    data.coords["col"].attrs["axis"] = "X"

    dims = CFConvention.read_spatial_dimensions(data)
    assert dims == ("row", "col")


def test_read_spatial_dimensions__cf_standard_name():
    """Test detecting dimensions from CF standard_name attributes."""
    data = xarray.DataArray(
        numpy.random.rand(10, 10),
        dims=["lat", "lon"],
        coords={
            "lat": ("lat", numpy.arange(10)),
            "lon": ("lon", numpy.arange(10)),
        },
    )
    data.coords["lat"].attrs["standard_name"] = "latitude"
    data.coords["lon"].attrs["standard_name"] = "longitude"

    dims = CFConvention.read_spatial_dimensions(data)
    assert dims == ("lat", "lon")


def test_read_spatial_dimensions__not_found():
    """Test that None is returned when spatial dimensions are not found."""
    data = xarray.DataArray(numpy.random.rand(10, 10), dims=["a", "b"])

    dims = CFConvention.read_spatial_dimensions(data)
    assert dims is None


def test_write_crs():
    """Test writing CRS to a DataArray."""
    data = xarray.DataArray(numpy.random.rand(10, 10), dims=["y", "x"])
    crs = CRS.from_epsg(4326)
    result = CFConvention.write_crs(data, crs=crs, grid_mapping_name="spatial_ref")

    assert "spatial_ref" in result.coords
    assert result.coords["spatial_ref"].attrs["spatial_ref"] == crs.to_wkt()
    assert result.coords["spatial_ref"].attrs["crs_wkt"] == crs.to_wkt()


def test_write_transform():
    """Test writing transform to a DataArray."""
    data = xarray.DataArray(numpy.random.rand(10, 10), dims=["y", "x"])
    transform = Affine(1.0, 0.0, 0.0, 0.0, -1.0, 10.0)

    result = CFConvention.write_transform(
        data, transform=transform, grid_mapping_name="spatial_ref"
    )

    assert "spatial_ref" in result.coords
    assert "GeoTransform" in result.coords["spatial_ref"].attrs
    assert (
        result.coords["spatial_ref"].attrs["GeoTransform"]
        == "0.0 1.0 0.0 10.0 0.0 -1.0"
    )



"""
This module contains helper methods for rasterio.crs.CRS.
"""
from typing import Any

import rasterio
import rasterio.crs
from packaging import version
from pyproj import CRS
from rasterio.errors import CRSError


def crs_from_user_input(crs_input: Any) -> rasterio.crs.CRS:
    """
    Return a rasterio.crs.CRS from user input.

    This is to deal with change in rasterio.crs.CRS
    as well as to assist in the transition between GDAL 2/3.

    Parameters
    ----------
    crs_input: Any
        Input to create a CRS.

    Returns
    -------
    rasterio.crs.CRS

    """
    if isinstance(crs_input, rasterio.crs.CRS):
        return crs_input
    try:
        # old versions of opendatacube CRS
        crs_input = crs_input.wkt
    except AttributeError:
        pass
    try:
        return rasterio.crs.CRS.from_user_input(crs_input)
    except CRSError:
        pass
    # use pyproj for edge cases
    crs = CRS.from_user_input(crs_input)
    if version.parse(rasterio.__gdal_version__) > version.parse("3.0.0"):
        return rasterio.crs.CRS.from_wkt(crs.to_wkt())
    return rasterio.crs.CRS.from_wkt(crs.to_wkt("WKT1_GDAL"))
